pro baselines_artemis, drifts_all, map_back, size_deg, maprot_deg=maprot_deg, $
               file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
               ind_aver=ind_aver, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
               galactic=galactic, time_1=time_1, t_proc=t_proc, force_nsig=force_nsig, $
               after_indiv=after_indiv, cube_mask=cube_mask, disjoint=disjoint


;;; determines one linear baseline per bolometer per subscan by robust fits
;;; (definition of subscan adapted to the multi-block case)

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

size_lim_nooff = 0.75
nsubs_crit = 4
ndev = 3.

dimxy = dimx * dimy

nb_row = 16L
nb_col = 18L
nb_mat = nb_row * nb_col
nmat = nb / nb_mat
if nmat eq 8L then nsub_x = 4L else nsub_x = 2L
nb_line = nsub_x * nb_row


ratio_thres = 0
med_wn = median(maxnoise)
med_wn_scans = fltarr(nscans)
for s = 0, nscans - 1 do begin
   wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1L)
   med_wn_scans(s) = median(maxnoise(wsubs, *))
endfor

restore, file_signal


nb_valid = total(total(finite(maxnoise), 1) ge nsubscans / 2L)
w = where(finite(signal) lt 1, cw)
mask_bolo_init = 1. + fltarr(nt, nb)
if cw ge 1 then mask_bolo_init(w) = 0.

norm_init = total(mask_bolo_init, 2)
w = where(norm_init lt nb_valid / 2., cw)
if cw ge 1 then norm_init(w) = 0.


restore, file_pixcoords
indxy = floor(x_bolos) + dimx * floor(y_bolos)
x_bolos = 0
y_bolos = 0

restore, file_times

restore, file_time_coarse

flag_save = intarr(nt_save)

tab_aver = lonarr(nt)
tab_aver(0:ind_aver(0)-1) = 0L
for i = 1L, nt_save - 1 do tab_aver(ind_aver(i-1):ind_aver(i)-1) = i


if (not keyword_set(map_back)) then map_back = fltarr(dimx, dimy)


tab_cdef = lonarr(nscans)
do_map_scans_nearest_artemis, map_scans, weight_scans, file_scalars=file_scalars, ind_scans=ind_scans, $
   ind_subscans=ind_subscans, maxnoise=maxnoise
map_scans = 0
for s = 0, nscans - 1 do begin
  w = where(weight_scans(*, *, s) gt 0., cdef)
  tab_cdef(s) = cdef
endfor
ind_sc_long = where(tab_cdef gt 3L * min(tab_cdef), cw)
if cw lt 1 then ind_sc_long = 0


;;; determine whether the field is covered by largely disjoint scans or not:

if (nscans gt 1 and (not keyword_set(force_nsig))) then begin
   w = where(weight_scans gt 0.)
   med_w_s = median(weight_scans(w))
   weight_cross_even = 1
   weight_cross_odd = 1
   for i = 0, nscans - 1 do begin
      if (i mod 2) eq 0 then weight_cross_even *= (weight_scans(*, *, i) / med_w_s) else weight_cross_odd *= (weight_scans(*, *, i) / med_w_s)
   endfor
   w = 0
   ind_even = where((indgen(nscans) mod 2) eq 0, n_even)
   ind_odd = where((indgen(nscans) mod 2) eq 1, n_odd)
   w_even = where(weight_cross_even gt 0., cwe)
   w_odd = where(weight_cross_odd gt 0., cwo)
   w_even = 0
   w_odd = 0
   if (cwe lt min(tab_cdef(ind_even)) / 4. or cwo lt min(tab_cdef(ind_odd)) / 4.) then disjoint = 1
endif
weight_scans = 0



restore, file_maps
w = where(map_weight gt 0., cw)
map_signal = 0
map_weight = 0
size_deg = sqrt(float(cw)) * pix_as / 3600.
print, '              size ~ ' + chain(size_deg, 2) + ' degree(s)'

if ((size_deg lt 1. and nscans le 20) or nt lt 50000L) then begin
   ns = 49
   if keyword_set(debug) then print, 'ns = 49'
endif else ns = 0



;;; subtract offsets at the start so that the average signal is not perturbed by missing data:
compute_offsets, file_scalars, off, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
   galactic=galactic
for s = 0, nscans - 1 do begin
   signal(ind_scans(s):ind_scans(s+1)-1, *) -= rebin(off(s, *), ind_scans(s+1)-ind_scans(s), nb)
   drifts_all(ind_sc_coarse(s):ind_sc_coarse(s+1)-1, *) += rebin(off(s, *), ind_sc_coarse(s+1)-ind_sc_coarse(s), nb)
endfor


if keyword_set(galactic) then iter_max = 1 else iter_max = 0
nloop = 3

for iter = 0, iter_max do begin


;;; loop to iteratively exclude sources from the fits:

for k = 0, nloop - 1 do begin

   mask_bolo = mask_bolo_init

   ;;; mask out bright sources flagged at previous iteration:
   if ((k gt 0 or iter gt 0 or keyword_set(force_nsig)) and keyword_set(galactic) $
    and max(abs(map_back), /nan) gt 0.) then begin
      w = where(finite(map_back(indxy)) lt 1, cw)
      if cw ge 1 then mask_bolo(w) = 0.
      w = 0
   endif


   norm = total(mask_bolo, 2)
   w = where(norm lt nb_valid / 2., cw)
   if cw ge 1 then norm(w) = 0.
   norm_denom = norm
   w = where(norm lt 0.5, cw)
   if cw ge 1 then norm_denom(w) = 1.
   aver = total(signal * mask_bolo, 2, /nan) / norm_denom
   if cw ge 1 then aver(w) = !values.f_nan



   ;;; baselines on scan length:

   for i = 0, nscans - 1 do begin

      wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans le ind_scans(i+1) - 1)
      med_noise = median(maxnoise(wsubs, *))

      nt_scan = ind_scans(i+1) - ind_scans(i)
      w1 = where(norm(ind_scans(i):ind_scans(i)+nt_scan/3L) gt 0., c1)
      w2 = where(norm(ind_scans(i)+2L*nt_scan/3L:ind_scans(i+1)-1L) gt 0., c2)
      if min([c1, c2]) ge max([(ind_scans(i+1) - ind_scans(i)) / 9L, 100L]) then begin
         wfit = where(norm(ind_scans(i):ind_scans(i+1)-1) gt 0., cw)
         time_ref = time_s(ind_scans(i))
         order = 1
         if (k le 1 and iter eq 0) then nomask = 1 else if tab_cdef(i) le 3. * min(tab_cdef) then nomask = 0
         if keyword_set(force_nsig) then nomask = 1
         fit_scan = absdev_fit_artemis(float(-time_ref + time_s(ind_scans(i)+wfit)), aver(ind_scans(i)+wfit), $
            coef=coef, ndev=ndev, /dyn, noise_level=med_noise, ns=ns, nomask=nomask)

         if keyword_set(debug) then begin
            time_2 = systime(/seconds)
            t_proc += (time_2 - time_1)
            wset, 0
            otherplot, time_s(ind_scans(i)+wfit), aver(ind_scans(i)+wfit), xtitle='time (s)', $
               ytitle='average signal', title='linear fit to scan '+chain(i)
            otherplot, /over, color=10, time_s(ind_scans(i)+wfit), fit_scan
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
            time_1 = systime(/seconds)
         endif

         fit_scan = coef(0)
         for pow = 1, order do fit_scan += (coef(pow) * float((-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1))^float(pow)))
         signal(ind_scans(i):ind_scans(i+1)-1, *) -= rebin(fit_scan, ind_scans(i+1)-ind_scans(i), nb)
         fit_coarse = coef(0)
         for pow = 1, order do fit_coarse += (coef(pow) * (-float(time_ref) + time_save(ind_sc_coarse(i):ind_sc_coarse(i+1)-1))^float(pow))
         drifts_all(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, *) += rebin(fit_coarse, ind_sc_coarse(i+1)-ind_sc_coarse(i), nb)
      endif
   endfor



   ;;; baselines on segments of 4 subscans:

   if ((k ge 1 or iter gt 0) and nsubscans / nscans ge 3 * nsubs_crit) then $
;;;   for mat = 0, nmat - 1 do begin
   for mat = 0, 0 do begin
;;;
;     nvis = 0
      drifts_1 = fltarr(nt_save)
      drifts_2 = fltarr(nt_save)
      drifts_base = fltarr(nt_save)
;;;      wmat = reform(rebin((mat mod nsub_x) * nb_row + lindgen(nb_row), nb_row, nb_col) $
;;;         + nb_line * (((mat / nsub_x) * nb_col + lindgen(nb_col)) ## (1L + lonarr(nb_row))), nb_mat)
      wmat = lindgen(nb)
      nb_mat = nb
;;;
      maxnoise_mat = maxnoise(*, wmat)

      norm = total(mask_bolo(*, wmat), 2)
      w = where(norm lt nb_valid / 2., cw)
      if cw ge 1 then norm(w) = 0.
      norm_denom = norm
      w = where(norm lt 0.5, cw)
      if cw ge 1 then norm_denom(w) = 1.
      aver = total(signal(*, wmat) * mask_bolo(*, wmat), 2, /nan) / norm_denom
      if cw ge 1 then aver(w) = !values.f_nan

      if total(finite(aver)) gt nt / 4L then for s = 0, nscans - 1 do begin

         order = 1
         wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans le ind_scans(s+1) - 1, csubs)
         med_noise = median(maxnoise_mat(wsubs, *))
;;
;;         nseg = round(csubs / float(nsubs_crit))
         nseg = round(csubs / float(nsubs_crit / 2L))
         if nseg gt 1 then begin
;;            nsubs_cut = max([2 * round(csubs / float(nseg) / 2.), 2])
;;            nseg = floor(csubs / float(nsubs_cut))
            nsubs_cut = nsubs_crit
            flag_1 = intarr(nt_save)
            flag_2 = intarr(nt_save)
;;            ind_start = ind_subscans(wsubs(nsubs_cut*lindgen(nseg)))
;;            ind_end = [ind_start(1:nseg-1) - 1, ind_scans(s+1) - 1]
            ind_start = [ind_scans(s), ind_subscans(wsubs(nsubs_cut/2L-1L+(nsubs_cut/2L)*lindgen(nseg-1L)))]
            ind_end = [ind_subscans(wsubs(nsubs_cut-1L+(nsubs_cut/2L)*lindgen(nseg-1L))) - 1L, ind_scans(s+1) - 1L]
            ind_1 = [ind_scans(s), ind_subscans(wsubs(nsubs_cut/2L+(nsubs_cut/2L)*lindgen(nseg-1L)))]
            ind_2 = [ind_subscans(wsubs(nsubs_cut-2L+(nsubs_cut/2L)*lindgen(nseg-1L))) - 1L, ind_scans(s+1) - 1L]
;;
            if (k le 1 and iter eq 0) then nomask = 1 else if tab_cdef(s) le 3. * min(tab_cdef) then nomask = 0
            if keyword_set(force_nsig) then nomask = 1

            for seg = 0L, nseg - 1L do begin
               nt_seg = ind_end(seg) - ind_start(seg) + 1
               nt_edge = nt_seg / nsubs_cut / 3L
               wfit = ind_start(seg) + lindgen(nt_seg)
               w1 = where(finite(aver(ind_start(seg):ind_start(seg)+nt_edge)) gt 0)
               w2 = where(finite(aver(ind_start(seg)+2L*nt_edge:ind_start(seg)+3L*nt_edge-1L)) gt 0)
               for subs = 1, nsubs_cut - 1 do begin
                  if (subs mod 2) eq 0 then begin
                     w1 = [w1, where(finite(aver(ind_start(seg)+subs*3L*nt_edge:ind_start(seg)+(subs*3L+1L)*nt_edge)) gt 0)]
                     w2 = [w2, where(finite(aver(ind_start(seg)+(subs*3L+2L)*nt_edge:ind_start(seg)+(subs*3L+3L)*nt_edge-1L)) gt 0)]
                  endif else begin
                     w1 = [w1, where(finite(aver(ind_start(seg)+(subs*3L+2L)*nt_edge:ind_start(seg)+(subs*3L+3L)*nt_edge-1L)) gt 0)]
                     w2 = [w2, where(finite(aver(ind_start(seg)+subs*3L*nt_edge:ind_start(seg)+(subs*3L+1L)*nt_edge)) gt 0)]
                  endelse
               endfor
               c1 = n_elements(w1)
               c2 = n_elements(w2)
               if min([c1, c2]) ge max([nt_seg / 12L, 100L]) then begin
                  wdef = where(finite(aver(wfit)) gt 0, cw)
                  time_ref = time_s(wfit(wdef(0)))
                  yfit = absdev_fit_artemis(float(-time_ref + time_s(wfit(wdef))), aver(wfit(wdef)), $
                     coef=coef, ndev=ndev, /dyn, noise_level=med_noise, ns=ns, nomask=nomask)

                  nt_seg_coarse = tab_aver(ind_end(seg)) - tab_aver(ind_start(seg)) + 1
                  fit_coarse = coef(0)
                  for pow = 1, order do fit_coarse += (coef(pow) * (-float(time_ref) + time_save(tab_aver(ind_start(seg)):tab_aver(ind_end(seg))))^float(pow))
;;                  drifts_base(tab_aver(ind_start(seg)):tab_aver(ind_end(seg))) += fit_coarse
                  if (seg mod 2L) eq 0 then drifts_1(tab_aver(ind_start(seg)):tab_aver(ind_end(seg))) += fit_coarse $
                     else drifts_2(tab_aver(ind_start(seg)):tab_aver(ind_end(seg))) += fit_coarse

                  ;;; eliminate discontinuities between successive segments:
;;                  flag_save(tab_aver(ind_start(seg))+1:tab_aver(ind_end(seg))-1) = 1
                  if tab_aver(ind_2(seg)) gt tab_aver(ind_1(seg)) + 2 then begin
                     if (seg mod 2L) eq 0 then flag_1(tab_aver(ind_1(seg))+1:tab_aver(ind_2(seg))-1) = 1 $
                      else flag_2(tab_aver(ind_1(seg))+1:tab_aver(ind_2(seg))-1) = 1
                  endif else begin
                     if (seg mod 2L) eq 0 then flag_1(tab_aver(ind_1(seg)):tab_aver(ind_2(seg))) = 1 $
                      else flag_2(tab_aver(ind_1(seg)):tab_aver(ind_2(seg))) = 1
                  endelse
               endif
            endfor

;;         wsave = where(flag_save eq 1, csave)
         time_ref = time_s(ind_scans(s))
;;         drifts_base(ind_sc_coarse(s)) = interpol([drifts_base(wsave(0)), drifts_base(wsave), drifts_base(wsave(csave-1))], $
;;            [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(s+1)-1) + timestep - time_ref)], $
;;            -float(time_ref) + time_save(ind_sc_coarse(s):ind_sc_coarse(s+1)-1))
         wsave = where(flag_1 eq 1 and abs(drifts_1) gt (Machar()).eps, csave)
         drifts_1(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L) = interpol([drifts_1(wsave(0)), drifts_1(wsave), drifts_1(wsave(csave-1))], $
            [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(s+1)-1) + timestep - time_ref)], $
            -float(time_ref) + time_save(ind_sc_coarse(s):ind_sc_coarse(s+1)-1))
         wsave = where(flag_2 eq 1 and abs(drifts_2) gt (Machar()).eps, csave)
         drifts_2(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L) = interpol([drifts_2(wsave(0)), drifts_2(wsave), drifts_2(wsave(csave-1))], $
            [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(s+1)-1) + timestep - time_ref)], $
            -float(time_ref) + time_save(ind_sc_coarse(s):ind_sc_coarse(s+1)-1))
         drifts_base(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L) = $
            (drifts_1(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L) + drifts_2(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L)) / 2.
;;;
         w2 = where(flag_2 eq 1)
         drifts_base(ind_sc_coarse(s):w2(0)-1L) = drifts_1(ind_sc_coarse(s):w2(0)-1L)
         w1 = where(flag_1 eq 1, c1)
         drifts_base(w1(c1-1L)+1L:ind_sc_coarse(s+1)-1L) = drifts_2(w1(c1-1L)+1L:ind_sc_coarse(s+1)-1L)
;;;;

;;;
         for subs = wsubs(0), wsubs(csubs-1) do begin
            drifts_base(tab_aver(ind_subscans(subs))) = !values.f_nan
            drifts_base(tab_aver(ind_subscans(subs+1)-1L)) = !values.f_nan
         endfor
         wdef = where(finite(drifts_base(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L)) gt 0, cdef, complement=wz)
         drifts_base(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L) = interpol([drifts_base(ind_sc_coarse(s)+wdef(0)), $
            drifts_base(ind_sc_coarse(s)+wdef), drifts_base(ind_sc_coarse(s)+wdef(cdef-1))], $
            [time_save(ind_sc_coarse(s)) - timestep, time_save(ind_sc_coarse(s)+wdef), time_save(ind_sc_coarse(s+1)-1) + timestep], $
            time_save(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L))
;;;

         if keyword_set(debug) then if s le 2 then begin
            wset, 0
            wdef = where(finite(aver(ind_scans(s):ind_scans(s+1)-1)) gt 0)
            otherplot, time_s(ind_scans(s)+wdef), aver(ind_scans(s)+wdef)
            otherplot, /over, color=10, time_save, drifts_base
            otherplot, /over, color=100, time_s(ind_scans(s))+[0.,0.], [-40.,40.]
            otherplot, /over, color=100, time_s(ind_scans(s+1)-1L)+[0.,0.], [-40.,40.]
            for subs = 0L, nsubscans - 1L do otherplot, /over, color=200, lstyle=2, time_s(ind_subscans(subs))+[0.,0.], [-10.,10.]
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         endif

         subs_drifts = interpol([drifts_base(ind_sc_coarse(s)), drifts_base(ind_sc_coarse(s):ind_sc_coarse(s+1)-1), $
            drifts_base(ind_sc_coarse(s+1)-1)], $
            [-float(timestep), -float(time_ref) + time_save(ind_sc_coarse(s):ind_sc_coarse(s+1)-1), $
            float(time_s(ind_scans(s+1)-1) + timestep - time_ref)], $
            float(-time_ref + time_s(ind_scans(s):ind_scans(s+1)-1)))
         signal(ind_scans(s):ind_scans(s+1)-1, wmat) -= rebin(subs_drifts, ind_scans(s+1)-ind_scans(s), nb_mat)
         flag_save *= 0

         endif

      endfor
      drifts_all(*, wmat) += rebin(drifts_base, nt_save, nb_mat)
   endfor



   ;;; baselines on subscan length:

   if k gt 0 then begin
      drifts_base = fltarr(nt_save, nb)

      for i = 0, nscans - 1 do begin
         wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans le ind_scans(i+1) - 1, csubs)
         med_noise = median(maxnoise(wsubs, *))
         ndev_subs = ndev

         for isubs = 0L, csubs - 1L do begin
            subs = wsubs(isubs)
            it0 = ind_subscans(subs)
            w = where(flag_nza(it0:ind_subscans(subs+1)-1) eq 1, cw)
            if (cw ge 400 and size_deg gt size_lim_nooff) then begin
               it0 += w(0)
               it1 = it0
               w = where(flag_nza(it1:ind_subscans(subs+1)-1) eq 0, cw)
               if cw eq 0 then it1 = ind_subscans(subs+1) - 1 else it1 += (w(0) - 1)
               it0_nom = it0
               it1_nom = it1
               it0 = max([ind_subscans(subs), it0 - 200])
               it1 = min([ind_subscans(subs+1) - 1, it1 + 200])
            endif else begin
               it0 = ind_subscans(subs)
               it1 = ind_subscans(subs+1) - 1
               it0_nom = it0
               it1_nom = it1
            endelse

            nt_fit = it1 - it0 + 1L
            if (size_deg gt 1. and ratio_thres lt 0.01 and nt_fit gt 1000) then begin
               wfit = [it0 + lindgen(nt_fit/3L), it1 - reverse(lindgen(nt_fit/3L))]
               nt_fit = 2L * (nt_fit / 3L)
            endif else wfit = it0 + lindgen(nt_fit)


            ;;; zero-order baselines:

            flux = signal(wfit, *)
            wz = where(mask_bolo(wfit, *) lt 0.5, cz)
            if cz ge 1 then flux(wz) = !values.f_nan
            base = drifts_base(tab_aver(it0):tab_aver(it1), *)
            subtract_offsets, file_scalars, flux, base, maxnoise_subs=reform(maxnoise(subs, *)), galactic=galactic
            drifts_base(tab_aver(it0):tab_aver(it1), *) = base


            ;;; first-order baselines (except for very structured fields):
;;;            if k eq nloop - 1 then begin
            if k gt 0 then begin
;;;

               mask_bolo_subs = mask_bolo(wfit, *)
;;;
               if (keyword_set(force_nsig) and (iter gt 0 or k ge 2)) then begin
                  nb_mat = nb_row * nb_col
                  use_nmat = nmat
               endif else begin
                  wmat = lindgen(nb)
                  nb_mat = nb
                  use_nmat = 1
                  use_nb_valid = nb_valid
               endelse
;;;
               for mat = 0, use_nmat - 1 do begin
                  if (keyword_set(force_nsig) and (iter gt 0 or k ge 2)) then begin
                     wmat = reform(rebin((mat mod nsub_x) * nb_row + lindgen(nb_row), nb_row, nb_col) $
                        + nb_line * (((mat / nsub_x) * nb_col + lindgen(nb_col)) ## (1L + lonarr(nb_row))), nb_mat)
                     use_nb_valid = total(finite(maxnoise(subs, wmat)))
                  endif

                  norm = total(mask_bolo_subs(*, wmat), 2)
                  w = where(norm lt use_nb_valid / 2., cw)
                  if cw ge 1 then norm(w) = 0.
                  norm_denom = norm
                  w = where(norm lt 0.5, cw)
                  if cw ge 1 then norm_denom(w) = 1.
                  aver = total(flux(*, wmat) * mask_bolo_subs(*, wmat), 2, /nan) / norm_denom
                  if cw ge 1 then aver(w) = !values.f_nan

                  w1 = where(finite(aver(0L:nt_fit/3L)) gt 0, c1)
                  w2 = where(finite(aver(2L*nt_fit/3L:*)) gt 0, c2)
                  if min([c1, c2]) ge nt_fit / 9L then begin
                     wdef = where(finite(aver) gt 0, cw)
                     time_ref = time_s(wfit(wdef(0)))
                     yfit = absdev_fit_artemis(float(-time_ref + time_s(wfit(wdef))), aver(wdef), coef=coef, $
                        ndev=ndev_subs, noise_level=med_noise, /dyn, ns=ns)
;;;
                     if (keyword_set(galactic) and (iter gt 0 or k ge 2) and tab_cdef(i) le 2L * min(tab_cdef)) then begin
                        wn = where(aver(wdef) - yfit lt (-med_noise), cn)
                        if cn ge cw / 30L then begin
                           ind_sort = sort(aver(wdef) - yfit)
                           aver(wdef(ind_sort(2L*cw/3L:cw-1L))) = !values.f_nan
                           wdef = where(finite(aver) gt 0, cw)
                           time_ref = time_s(wfit(wdef(0)))
                           yfit = absdev_fit_artemis(float(-time_ref + time_s(wfit(wdef))), aver(wdef), coef=coef, $
                              ndev=ndev_subs, noise_level=med_noise, /dyn, ns=ns)
                        endif
                     endif
;;;
                     fit_subscan = coef(0) + coef(1) * float(-time_ref + time_s(wfit))
                     flux(*, wmat) -= rebin(fit_subscan, nt_fit, nb_mat)
                     drifts_base(tab_aver(it0):tab_aver(it1), wmat) += rebin(coef(0) + coef(1) * $
                        (-float(time_ref) + time_save(tab_aver(it0):tab_aver(it1))), $
                        tab_aver(it1)-tab_aver(it0)+1, nb_mat)

                  endif
               endfor
            endif

            ;;; eliminate discontinuities between successive scan legs:
            if tab_aver(it1) gt tab_aver(it0) + 2 then flag_save(tab_aver(it0)+1:tab_aver(it1)-1) = 1 $
               else flag_save(tab_aver(it0):tab_aver(it1)) = 1
         endfor


;;;         wsave = where(flag_save eq 1, csave)
         time_ref = time_s(ind_scans(i))
         for b = 0L, nb - 1L do if total(finite(maxnoise(wsubs, b)), 1) gt 0 then begin
;;;
            wsave = where(flag_save eq 1 and abs(drifts_base(*, b)) gt (Machar()).eps, csave)
            if csave ge (ind_sc_coarse(i+1) - ind_sc_coarse(i)) / 10L then begin
;;;
            drifts_base(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, b) = interpol([drifts_base(wsave(0), b), $
               drifts_base(wsave, b), drifts_base(wsave(csave-1), b)], $
               [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)], $
               -float(time_ref) + time_save(ind_sc_coarse(i):ind_sc_coarse(i+1)-1))
            subs_drifts = interpol([drifts_base(ind_sc_coarse(i), b), drifts_base(ind_sc_coarse(i):ind_sc_coarse(i+1)-1, b), $
               drifts_base(ind_sc_coarse(i+1)-1, b)], $
               [-float(timestep), -float(time_ref) + time_save(ind_sc_coarse(i):ind_sc_coarse(i+1)-1), $
               float(time_s(ind_scans(i+1)-1) + timestep - time_ref)], $
               float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))

;;;
if keyword_set(debug) then if (i eq 0 and b ge 100 and b le 105) then begin
   wset, 0
   otherplot, time_save(tab_aver(ind_scans(i)):tab_aver(ind_scans(i+1))-1), flag_save(tab_aver(ind_scans(i)):tab_aver(ind_scans(i+1))-1)
   for jj = 0L, nsubscans - 1L do otherplot, /over, color=200, lstyle=2, time_s(ind_subscans(jj))+[0.,0.], [-10.,10.]
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   w = where(finite(signal(ind_scans(i):ind_scans(i+1)-1, b)) gt 0)
   otherplot, time_s(ind_scans(i)+w), signal(ind_scans(i)+w, b)
   otherplot, /over, color=10, time_s(ind_scans(i):ind_scans(i+1)-1), subs_drifts
   for jj = 0L, nsubscans - 1L do otherplot, /over, color=200, lstyle=2, time_s(ind_subscans(jj))+[0.,0.], [-10.,10.]
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
endif
;;;

            signal(ind_scans(i):ind_scans(i+1)-1, b) -= subs_drifts
;;;
            endif
;;;
         endif

         flag_save *= 0
      endfor

      drifts_all += drifts_base
   endif
   mask_bolo = 0


   save, file=file_signal, signal, /xdr





   ;;; produce source mask for next iteration:

   if keyword_set(galactic) then $
    if ((iter lt iter_max and k gt 0) or (iter eq iter_max and k lt nloop - 1 and k ge nloop - 2)) then begin

      proj_map_nearest_artemis, file_scalars=file_scalars, maxnoise=maxnoise, $
        ind_subscans=ind_subscans, /comp_errors
      restore, file_maps

      w = where(map_weight le 0.)
      map_error(w) = !values.f_nan

      ;;; set approximately the zero level of the map:
      stat_map_artemis, med, sig, map=map_signal, weightmap=map_weight, noise_level=med_wn, use_map_back=map_back
;;;
      map_med = med + fltarr(dimx, dimy)
      map_sig = sig + fltarr(dimx, dimy)
      w = where(map_weight le 0.)
      map_med(w) = !values.f_nan
      map_sig(w) = !values.f_nan
      if keyword_set(disjoint) then do_map_scans_nearest_artemis, map_scans, weight_scans, file_scalars=file_scalars, $
         ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, /comp_errors, error_scans $
       else do_map_scans_nearest_artemis, map_scans, weight_scans, file_scalars=file_scalars, ind_scans=ind_scans, $
         ind_subscans=ind_subscans, maxnoise=maxnoise
      tab_med_sc = fltarr(nscans)
      tab_sig_sc = fltarr(nscans)
      for s = 0, nscans - 1 do begin
         weightmap = weight_scans(*, *, s)
         stat_map_artemis, med_sc, sig_sc, map=map_scans(*, *, s), weightmap=weightmap, noise_level=med_wn_scans(s)
         tab_med_sc(s) = med_sc
         tab_sig_sc(s) = sig_sc
      endfor
      weightmap = 0
;;;

;;;
      if keyword_set(disjoint) then if (not keyword_set(after_indiv) and ((not keyword_set(force_nsig)) or iter lt 1)) then begin
         for s = 0, nscans - 1 do map_scans(*, *, s) -= tab_med_sc(s)
         if nscans gt 1 then map = total(map_scans * weight_scans, 3, /nan) / total(weight_scans, 3, /nan) else map = map_scans
         if keyword_set(debug) then begin
            disp_ima, win=20, map_signal, weightmap=map_weight, title='map_signal'
            disp_ima, win=21, map, weightmap=map_weight, title='map with scan median offsets'
            stop
         endif
         map_signal = map
      endif
;;;



;;;if (iter lt iter_max or k lt nloop - 1) then begin
ws = where(tab_cdef gt 3L * min(tab_cdef), cs)
if cs ge 1 then begin
   if cs eq 1 then map_count = float(weight_scans(*, *, ws) gt 0.) else map_count = total(weight_scans(*, *, ws) gt 0., 3)
   w = where(map_count gt 0.5)
   map_med(w) = 0.
   map_sig(w) = 0.
   map_norm = fltarr(dimx, dimy)
   for is = 0, cs - 1 do begin
      s = ws(is)
      w = where(weight_scans(*, *, s) gt 0.)
      map_norm(w) += weight_scans(w+s*dimxy)
      map_med(w) += tab_med_sc(s) * weight_scans(w+s*dimxy)
      map_sig(w) += tab_sig_sc(s)^2. * weight_scans(w+s*dimxy)
   endfor
   w = where(map_count gt 0.5)
   map_med(w) /= map_norm(w)
   map_sig(w) = sqrt(map_sig(w) / map_norm(w))
endif

ws = where(tab_cdef le 3L * min(tab_cdef), cs)
if cs eq 1 then map_count = float(weight_scans(*, *, ws) gt 0.) else map_count = total(weight_scans(*, *, ws) gt 0., 3)
w = where(map_count gt 1.5, cw)
if cw ge 1 then begin
   map_med(w) = 0.
   map_sig(w) = 0.
   map_norm = fltarr(dimx, dimy)
   for is = 0, cs - 1 do begin
      s = ws(is)
      w = where(weight_scans(*, *, s) gt 0. and map_count gt 1.5)
      map_norm(w) += weight_scans(w+s*dimxy)
      map_med(w) += tab_med_sc(s) * weight_scans(w+s*dimxy)
      map_sig(w) += tab_sig_sc(s)^2. * weight_scans(w+s*dimxy)
   endfor
   w = where(map_count gt 1.5)
   map_med(w) /= map_norm(w)
   map_sig(w) = sqrt(map_sig(w) / map_norm(w))
endif

;;;
w = where(map_weight gt 0.)
map_med(w) = median(map_med(w))
map_sig(w) = median(map_sig(w))
;;;

      w = where(map_weight gt 0.)
      map_signal(w) -= map_med(w)
      map_med = 0

      map_back = map_signal
      map_back_rot = map_back
      map_error_rot = map_error
      w = where(map_weight gt max(map_weight) / 10.)
      med_weight = median(map_weight(w))
      sig_weight = sqrt(mean((map_weight(w) - med_weight)^2.))
      w = where(map_weight gt max(map_weight) / 10. and abs(map_weight - med_weight) lt 2.5 * sig_weight)
      sig_weight = sqrt(mean((map_weight(w) - med_weight)^2.))
      thres_weight_init = max([med_weight - 2.5 * sig_weight, 0.1 * med_weight])
      thres_weight = min([0.6 * med_weight, thres_weight_init])
      if (not keyword_set(disjoint)) then begin
         w = where(map_weight gt max(map_weight) / 4. and abs(map_weight - med_weight) lt 1.5 * sig_weight)
         sig_weight = sqrt(mean((map_weight(w) - med_weight)^2.))
         thres_weight = med_weight - min([max(map_weight) - med_weight, 2. * sig_weight])
         thres_weight = max([0.4 * med_weight, thres_weight])
      endif


      ;;; try to produce criterion to define the central zone better adapted to weight maps of any shape:
      if (not keyword_set(dist_edge)) then begin
         w = where(map_weight gt 0.)
         xm = median(w mod dimx, /even)
         ym = median(w / dimx, /even)
         map_dist = sqrt(rebin((lindgen(dimx) - xm)^2., dimx, dimy) + $
            ((lindgen(dimy) - ym)^2. ## (1. + fltarr(dimx))))
         w = where(map_weight gt thres_weight, np_tot, complement=wz)
         map_dist(wz) = !values.f_nan
         map_dist_2 = map_dist
         w = where(map_weight gt thres_weight + median(map_weight(where(map_weight gt 0.))) / 10.)
         map_dist_2(w) = !values.f_nan
         dist_edge = median(map_dist_2, /even)
         if keyword_set(visu) then begin
            disp_ima, win=28, map_dist, min_map=0.4*dist_edge, title='zone used to quantify max. brightness'
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
            wdelete, 28
         endif
      endif


      w = where(map_weight le 0.)
      map_back_rot(w) = !values.f_nan
      map_error_rot(w) = !values.f_nan
      inimap_back_rot = map_back_rot
      if keyword_set(visu) then inimap_error_rot = map_error_rot
      w = where(map_weight le thres_weight)
      map_back_rot(w) = !values.f_nan
      map_error_rot(w) = !values.f_nan
      if abs(maprot_deg) gt 1.e-4 then begin
         map_back_rot = rot(map_back_rot, maprot_deg, missing=!values.f_nan)
         map_error_rot = rot(map_error_rot, maprot_deg, missing=!values.f_nan)
         inimap_back_rot = rot(inimap_back_rot, maprot_deg, missing=!values.f_nan)
         if keyword_set(visu) then inimap_error_rot = rot(inimap_error_rot, maprot_deg, missing=!values.f_nan)
      endif
      dimx_rot = n_elements(map_back_rot(*, 0))
      dimy_rot = n_elements(map_back_rot(0, *))

      w = where(finite(inimap_back_rot) gt 0., np_tot)
      wx = w mod dimx_rot
      wy = w / dimx_rot
      wx_min = min(wx)
      wx_max = max(wx)
      wy_min = min(wy)
      wy_max = max(wy)


      if (not keyword_set(force_nsig)) then begin
         nerr = 10.
         nsig = 2.5
      endif
      if keyword_set(force_nsig) then begin
         nerr = 30.
         nsig = 3.
         if k gt 1 then begin
            nerr = 20.
            nsig = 2.5
         endif
         if iter gt 0 then begin
            nerr = 15.
            nsig = 2.2
         endif
      endif

;;;
if keyword_set(visu) then begin
   if (k eq 0 and (not keyword_set(force_nsig))) then disp_ima, win=26, map_weight, title='weight map'
   print, 'thres_weight = ' + chain(thres_weight)
   disp_ima, win=27, smooth(inimap_back_rot, 5, /nan) / smooth(map_sig, 5, /nan), title='map / sig', min_map=-5., max_map=nsig
   disp_ima, win=28, smooth(inimap_back_rot, 5, /nan) / smooth(inimap_error_rot, 5, /nan), $
      title='map / error map', min_map=-6., max_map=nerr
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   if (k eq 0 and (not keyword_set(force_nsig))) then wdelete, 26
   wdelete, 27, 28
   inimap_error_rot = 0
endif
inimap_back_rot = 0
;;;

      map_sm = smooth(map_back_rot, 5, /nan)
      map_sig_sm = smooth(map_sig, 5, /nan)
      map_sig = 0
      w = where(finite(map_back_rot) gt 0. and map_sm gt nsig * map_sig_sm $
         and map_sm gt (nsig - 1.) * map_sig_sm and map_sm gt nerr * smooth(map_error_rot, 5, /nan), cw)

      wmid_init = where(map_dist le 0.4 * dist_edge, cmid_init)
      if cmid_init gt 0.5 * np_tot then begin
         wx = w mod dimx_rot
         wy = w / dimx_rot
         wmid = where(wx ge (2*wx_min+wx_max)/3 and wx le (wx_min+2*wx_max)/3 $
            and wy ge (2*wy_min+wy_max)/3 and wy le (wy_min+2*wy_max)/3, cmid)
      endif else wmid = where(map_dist(w) le 0.4 * dist_edge, cmid)
      if cw ge 1 then ratio_mid = float(cmid) / float(cw) else ratio_mid = 1.
      ratio_thres = float(cw) / float(np_tot)
      ratio_thres_init = ratio_thres


      if (k eq nloop - 2 and keyword_set(debug)) then $
         print, '              cmid / cw = ' + chain(ratio_mid, 2) + ' and cw / np_tot = ' + chain(ratio_thres, 2)


      while (((ratio_mid lt 0.1 and ratio_thres gt 0.1) $
         or (ratio_mid lt 0.5 and ratio_thres gt 0.2) or (ratio_mid / ratio_thres lt 2.5 and ratio_thres gt 0.1)) $
         or ratio_thres gt 0.5 or (ratio_mid lt 0.2 - (ratio_thres - 0.3) / 2. and ratio_thres gt 0.3)) do begin

         nsig += 0.5
         w = where(finite(map_back_rot) gt 0. and map_sm gt nsig * map_sig_sm $
            and map_sm gt (nsig - 1.) * map_sig_sm and map_sm gt nerr * smooth(map_error_rot, 5, /nan), cw)
         if cmid_init gt 0.5 * np_tot then begin
            wx = w mod dimx_rot
            wy = w / dimx_rot
            wmid = where(wx ge (2*wx_min+wx_max)/3 and wx le (wx_min+2*wx_max)/3 $
               and wy ge (2*wy_min+wy_max)/3 and wy le (wy_min+2*wy_max)/3, cmid)
         endif else wmid = where(map_dist(w) le 0.4 * dist_edge, cmid)
         if cw ge 1 then ratio_mid = float(cmid) / float(cw) else ratio_mid = 1.
         ratio_thres = float(cw) / float(np_tot)
      endwhile


      if cw ge 1 then begin
         map_sm =  smooth(map_back, 5, /nan)
         w = where(((map_weight gt thres_weight and map_sm gt nsig * map_sig_sm $
            and map_sm gt (nsig - 1.) * map_sig_sm) or map_sm gt 6. * map_sig_sm) $
            and map_sm gt nerr * smooth(map_error, 5, /nan), cw)
         if cw ge 1 then map_back(w) = !values.f_nan
      endif


;;;
      if (not keyword_set(after_indiv) and keyword_set(disjoint)) then begin
      stat_map_artemis, med_back, sig_back, map=map_back, weightmap=map_weight, noise_level=med_wn
      if (keyword_set(galactic) and (k ge 1 or iter gt 0)) then begin
         tab_cz = lonarr(nscans)
         for s = 0, nscans - 1 do begin
            w = where(finite(map_back) lt 1 and weight_scans(*, *, s) gt 0., cz)
            tab_cz(s) = cz
         endfor
;;
         if (not keyword_set(force_nsig)) then for s = 0, nscans - 1 do if tab_cdef(s) gt 3L * min(tab_cdef) then begin
            w = where(map_weight gt thres_weight and weight_scans(*, *, s) gt 0. $
               and weight_scans(*, *, s) lt max(weight_scans(*, *, s) / 2.) and finite(map_back) lt 1, cw)
            w0 = where(map_weight gt thres_weight and weight_scans(*, *, s) gt 0. and weight_scans(*, *, s) lt max(weight_scans(*, *, s) / 2.), c0)
            if cw gt 0.2 * c0 then map_back(w) = 0.
         endif
;;
         for s = 0, nscans - 1 do if tab_cdef(s) le 3L * min(tab_cdef) then begin
            stat_map_artemis, med_sc, sig_sc, map=map_scans(*, *, s), weightmap=weight_scans(*, *, s), $
               noise_level=med_wn_scans(s)
            map_sm = smooth(map_scans(*, *, s), 5, /nan)
            w = where(((map_weight gt thres_weight and map_scans(*, *, s) gt med_sc + nsig * sig_back $
               and map_sm gt med_sc + (nsig - 1.) * sig_back) or map_scans(*, *, s) gt med_sc + 6. * sig_back) $
               and map_sm gt med_sc + (nerr - 1.) * smooth(error_scans(*, *, s), 5, /nan) $
               and map_sm gt med_sc + nerr * smooth(map_error, 5, /nan) $
               and weight_scans(*, *, s) gt 0., cw)
            if cw ge 1 then map_back(w) = !values.f_nan
         endif
      endif
      endif
;;;


      if keyword_set(visu) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         print, 'nsig = ' + chain(nsig, 1)
         weight_back = map_weight
         w = where(finite(map_back) lt 1)
         weight_back(w) = 0.
         disp_ima, win=10, map_back, weightmap=weight_back, title='source mask for baseline fits'
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         time_1 = systime(/seconds)
      endif
   endif

   map_scans = 0
   weight_scans = 0
   error_scans = 0
   map_back_rot = 0
   map_error_rot = 0
   map_sm = 0
   map_sig_sm = 0
   map_signal = 0
   map_error = 0

endfor


if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   wset, 0
   otherplot, time_save, drifts_all(*, bref), charsize=2., xtitle='coarse time (s)', $
      title='baseline for bolometer ' + chain(bref)
   for i = 0, nscans - 1 do begin
      y = [min(drifts_all, /nan), max(drifts_all, /nan)]
      otherplot, /over, lstyle=2, color=200, time_s(ind_scans(i))+[0., 0.], y
      otherplot, /over, lstyle=2, color=200, time_s(ind_scans(i+1)-1)+[0., 0.], y
   endfor
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif


endfor

mask_bolo_init = 0


if keyword_set(disjoint) then begin
   do_map_scans_nearest_artemis, map_scans, weight_scans, file_scalars=file_scalars, ind_scans=ind_scans, $
      ind_subscans=ind_subscans, maxnoise=maxnoise
   if (not keyword_set(force_nsig)) then for s = 0, nscans - 1 do begin
      stat_map_artemis, med_sc, sig_sc, map=map_scans(*, *, s), weightmap=weight_scans(*, *, s), noise_level=med_wn_scans(s)
      signal(ind_scans(s):ind_scans(s+1)-1L, *) -= med_sc
      drifts_all(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L, *) += med_sc
   endfor else begin
      w = where(tab_cdef le 3L * min(tab_cdef))
      weight_ref = total(weight_scans(*, *, w), 3)
      map_ref = total(map_scans(*, *, w) * weight_scans(*, *, w), 3, /nan) / weight_ref
      wz = where(finite(map_back) lt 1, cz)
      if cz ge 1 then weight_ref(wz) = 0.
      stat_map_artemis, med_ref, sig_ref, map=map_ref, weightmap=weight_ref, noise_level=med_wn
      w = where(tab_cdef gt 3L * min(tab_cdef), cw)
      for is = 0, cw - 1 do begin
         s = w(is)
         weight_s = weight_scans(*, *, s)
         if cz ge 1 then weight_ref(wz) = 0.
         stat_map_artemis, med_sc, sig_sc, map=map_scans(*, *, s), weightmap=weight_s, $
            noise_level=med_wn_scans(s)
         signal(ind_scans(s):ind_scans(s+1)-1L, *) -= (med_sc - med_ref)
         drifts_all(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L, *) += (med_sc - med_ref)
      endfor
   endelse
endif



if keyword_set(cube_mask) then begin
   cube_mask = fltarr(nt, nb)
   w = where(finite(map_back(indxy)) lt 1, cw)
   if cw ge 1 then cube_mask(w) = 1.
   w = 0
endif



save, file=file_signal, signal, /xdr
print, ''


end
