pro compute_indiv_drifts_artemis, array_drifts, array_npass, w_glitches, nb_glitch, $
                          maxnoise=maxnoise, highfnoise=highfnoise, ind_scans=ind_scans, ind_aver=ind_aver, $
                          file_scalars=file_scalars, min_nsamp=min_nsamp, detect_gl=detect_gl, $
                          t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, time_1=time_1, t_proc=t_proc, $
                          iter_indiv=iter_indiv, n_lt_indiv=n_lt_indiv, ind_subscans=ind_subscans


;;; computes the individual drift of each bolometer by using the redundancy,
;;; after the average drift has been subtracted
;;; switches to a finer grid or drops the computation whenever significant flux variations are detected
;;; (thus avoids contamination from compact sources, strong brightness gradients, and glitches)
;;; performs second-pass deglitching

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_pixcoords

restore, file_signal


;;; to correct for slight pointing differences of the samples within the coarse pixel,
;;; using the map produced at the previous iteration

restore, file_maps

map_error = 2.5 / sqrt(map_weight)
med_noise = median(maxnoise)

stat_map_artemis, med_map, sig_map, map=map_signal, weightmap=map_weight, noise_level=med_noise

if keyword_set(map_weight_drifts) then map_weight = map_weight_drifts
w = where(map_weight gt 0.)
med = median(map_weight(w))
h = histogram(map_weight(w), min=med/10., binsize=10., locations=index)
wh = where(h gt max(h) / 5.)
w = where(map_weight gt 0. and map_weight le index(max(wh)))
med = median(map_weight(w))
w = where(map_weight gt 0.3 * med and map_weight le index(max(wh)))
med = median(map_weight(w))
map_weight /= med
w = where(map_weight gt 1.)
map_weight(w) = 1.
w = where((map_weight gt 0. and map_weight lt 0.1) or map_weight lt 0.)
map_weight(w) = 0.1


thres_neigh = 22.
thres_diff = 30.


;;;fact_bin = 5L
fact_bin = 3L
;;;
map_rebin = rebin(map_signal, fact_bin*dimx, fact_bin*dimy)
diff_signals = signal
for b = 0L, nb - 1 do diff_signals(*, b) -= $
   map_rebin(floor(fact_bin*x_bolos(*, b)) + fact_bin*dimx * floor(fact_bin*y_bolos(*, b)))
map_rebin = 0


restore, file_times
flag_nza = 0

tab_aver = lonarr(nt)
tab_aver(0:ind_aver(0)-1) = 0L
for i = 1L, nt_save - 1 do tab_aver(ind_aver(i-1):ind_aver(i)-1) = i

tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]


restore, file_projcoarse

;;; coarse map to visualize the drift correction type:
map_proj = fltarr(dimx_iter, dimy_iter)
dimxy_iter = dimx_iter * dimy_iter
map_proj(wproj mod dimxy_iter) = 2
map_pointcorr = fltarr(dimx_iter, dimy_iter)
map_pointcorr(wproj mod dimxy_iter) = 2


restore, file_ntmin_pass

w = where(ntmin_pass lt 3. * median(ntmin_pass))
med_ntmin_pass = median(ntmin_pass(w))
if med_ntmin_pass ge min_nsamp * fact then begin
   flag_fine = 1
   ntmin_pass_fine = ntmin_pass / fact
   w = where(ntmin_pass_fine lt min_nsamp, cw)
   if cw ge 1 then ntmin_pass_fine(w) = min_nsamp
   nproj_zoom = fact^2
   tab_dxpf = indgen(nproj_zoom) mod fact
   tab_dypf = indgen(nproj_zoom) / fact
endif else begin
   flag_fine = 0
   x_bolos = 0
   y_bolos = 0
endelse
;;;w = where(ntmin_pass lt min_nsamp, cw)
;;;if cw ge 1 then ntmin_pass(w) = min_nsamp
w = where(ntmin_pass lt 5, cw)
if cw ge 1 then ntmin_pass(w) = 5
;;;



array_drifts = fltarr(nt_save, nb)
array_weight = fltarr(nt_save, nb)
array_npass = fltarr(nt_save, nb)



w_glitches = [-9L]
ncrit_gl = max([3, npass_crit])
fact_cen_gl = 0.75
if f_fwhm gt 2. then fact_cen_gl += (0.1 * (f_fwhm - 2.))
if inst eq 'spire' then thres_oscill = 3. else thres_oscill = 5.
thres_gl = 5.
thres_around = 5.
if inst eq 'spire' then thres_det = 4. else thres_det = 5.
dt_sig_gl = 4. * t_cross_fwhm
n_visu_gl = 0L


save_npass_save = -9L + lonarr(nproj)


ind = sort(signal)
med_all = median(signal(ind(0:2*nt*nb/3)))
ind = 0


sq_fact = fact^2.



off = floor(0.5 * fact) * (wproj / dimxy_iter)
x0 = off + ((wproj mod dimxy_iter) mod dimx_iter) * fact
y0 = off + ((wproj mod dimxy_iter) / dimx_iter) * fact
off = 0

;;;
;map_cp = fltarr(dimx_iter, dimy_iter)
;;;
for k = 0L, nproj - 1L do begin

   ;;; quicker way of evaluating: wp = where(indxy_beam_coarse eq wproj(k), cp):
   revind_0 = revind(wproj(k))
   revind_1 = revind(wproj(k)+1)
   cp = revind_1 - revind_0
;;;
;if wproj(k) lt dimxy_iter then map_cp(wproj(k)) = cp
;;;
   wp = revind(revind_0:revind_1-1L)
   wt = wp mod nt

;;;   if cp ge median(ntmin_pass(wt)) then begin
   if cp ge min_nsamp then begin
;;;
      wbolo = wp / nt
      distrib_time = time_s(wt)

      ;;; find the start indices of each distinct bolometer passage through the pixel:
      diff_bolo = [1L, wbolo(1:*) - wbolo(0:cp-2)]
      diff_time = [0.D0, distrib_time(1:*) - distrib_time(0:cp-2)]
      ind_startpass = where(diff_bolo ge 1L or diff_time gt maxtimediff, npass)
      if npass ge 3 then begin
         ind_endpass = [ind_startpass(1:*) - 1L, cp-1L]
         ws = tab_subscan(wt(ind_startpass))
         med_noise = mean(maxnoise(ws, wbolo(ind_startpass)))

         ;;; correct for pointing differences using the current map:
         x1 = min([x0(k) + fact - 1, dimx - 1])
         y1 = min([y0(k) + fact - 1, dimy - 1])
         norm_npass_cen = mean(map_weight(x0(k):x1, y0(k):y1))
         box = map_signal(x0(k):x1, y0(k):y1)
         box_error = map_error(x0(k):x1, y0(k):y1)
         w = where(finite(box_error) gt 0, cw)
         if cw ge 3 then begin
            x2 = max([x0(k) - 3 * fact, 0])
            x3 = min([x0(k) + 4 * fact - 1, dimx - 1])
            y2 = max([y0(k) - 3 * fact, 0])
            y3 = min([y0(k) + 4 * fact - 1, dimy - 1])
            box_ar = map_signal(x2:x3, y2:y3)
            box_error_ar = map_error(x2:x3, y2:y3)
            w_ar = where(finite(box_error_ar) gt 0)
            mean_neigh = mean((box_ar(w_ar) - med_map) / (med_noise * box_error_ar(w_ar)))
            x2 = max([x0(k) - 2 * fact, 0])
            x3 = min([x0(k) + 3 * fact - 1, dimx - 1])
            y2 = max([y0(k) - 2 * fact, 0])
            y3 = min([y0(k) + 3 * fact - 1, dimy - 1])
            box_ar = map_signal(x2:x3, y2:y3)
            box_error_ar = map_error(x2:x3, y2:y3)
            box_error_ar(x0(k)-x2:x1-x2, y0(k)-y2:y1-y2) = !values.f_nan
            w_ar = where(finite(box_error_ar) gt 0)
            diff_box = max((box(w) - median(box_ar(w_ar), /even)) / (med_noise * box_error(w)), /absolute)
         endif else begin
            mean_neigh = 0.
            diff_box = 0.
         endelse
         distrib_flux = diff_signals(wp)
         distrib_true_flux = signal(wp)

         ;;; to detect whether the flux variations within the pixel require the finer grid:
         ns_pass = ind_endpass - ind_startpass + 1
         flag_ntmin = intarr(npass)

         tab_ntmin = intarr(npass)
         for i_ntmin = 0L, npass - 1L do $
            tab_ntmin(i_ntmin) = median(ntmin_pass(wt(ind_startpass(i_ntmin):ind_endpass(i_ntmin))), /even)
         wcen = where(ns_pass ge tab_ntmin, npass_cen)
         if npass_cen ge 3 then begin
            flag_ntmin(wcen) = 1
            hfnoise_pass = highfnoise(ws, wbolo(ind_startpass))
            maxnoise_pass = maxnoise(ws, wbolo(ind_startpass))


            ;;; reduce execution time by looping on maximum number of samples per crossing
            ;;; instead of looping on number of crossings (which scales with number of bolometers):

            ns_max = max(ns_pass)
            if ns_max ge 3 then begin
               tab_flux = fltarr(ns_max, npass) + !values.f_nan
               tab_true_flux = fltarr(ns_max, npass) + !values.f_nan
               for ind_s = 0, ns_max - 1 do begin
                  w = where(ns_pass gt ind_s)
                  tab_flux(ind_s, w) = distrib_flux(ind_startpass(w)+ind_s)
                  tab_true_flux(ind_s, w) = distrib_true_flux(ind_startpass(w)+ind_s)
               endfor
               med_pass = total(tab_flux, 1, /nan) / float(ns_pass)
               true_med_pass = total(tab_true_flux, 1, /nan) / float(ns_pass)
               var_pass = total(abs(tab_flux - (med_pass ## (fltarr(ns_max) + 1.))), 1, /nan) / float(ns_pass)
               true_var_pass = total(abs(tab_true_flux - (true_med_pass ## (fltarr(ns_max) + 1.))), 1, /nan) / float(ns_pass)
               var_oscill_pass = sqrt(total((tab_flux(1:ns_max-1, *) - tab_flux(0:ns_max-2, *))^2., 1, /nan) / float(ns_pass - 1))
               sq_maxnoise_pass = maxnoise_pass^2.

               pointcorr = 1
               if (mean_neigh * norm_npass_cen le thres_neigh and diff_box * norm_npass_cen le thres_diff) then begin
                  pointcorr = 0
                  med_pass = true_med_pass
                  var_pass = true_var_pass
                  var_oscill_pass = sqrt(total((tab_true_flux(1:ns_max-1, *) - tab_true_flux(0:ns_max-2, *))^2., 1, /nan) / float(ns_pass - 1))
                  map_pointcorr(wproj(k) mod dimxy_iter) = 1
               endif

               if pointcorr eq 1 then $
                  wsave = where(flag_ntmin eq 1 and true_var_pass lt hfnoise_pass and var_pass lt hfnoise_pass, npass_save) $
                  else wsave = where(flag_ntmin eq 1 and (true_var_pass lt hfnoise_pass or var_pass lt hfnoise_pass), npass_save)
;;;
               if ((mean_neigh * norm_npass_cen le 12. and diff_box * norm_npass_cen le 12.) $
                  or diff_box * norm_npass_cen le 3.) then diffuse = 1 else diffuse = 0
;;;


            ;;; glitch detection:
            if keyword_set(detect_gl) then if (npass_save ge ncrit_gl and npass_save ge $
               floor(fact_cen_gl * npass_cen)) then begin
               med_ref = median(true_med_pass(wsave), /even) - med_all
               wgl = where(flag_ntmin eq 1 and var_oscill_pass gt med_ref $
                  and var_oscill_pass gt thres_oscill * hfnoise_pass, npass_gl)

               for i = 0L, npass_gl - 1L do begin
                  ib = wbolo(ind_startpass(wgl(i)))
                  it = wt((ind_startpass(wgl(i)) + ind_endpass(wgl(i))) / 2)
                  delta_t = abs(time_s - time_s(it))
                  dt_det_gl = ns_pass(wgl(i)) * timestep
                  w = where(delta_t le dt_sig_gl)
                  med_gl = median(signal(w, ib))
                  sig_gl = sqrt(median((signal(w, ib) - med_gl)^2.))
                  dev = max([hfnoise_pass(wgl(i)), sig_gl])
                  w = wt(ind_startpass(wgl(i)):ind_endpass(wgl(i)))
                  peak = max(abs(signal(w, ib) - med_gl), it)
                  it = w(it)
                  delta_t = abs(time_s - time_s(it))
                  w = where(delta_t le dt_det_gl and delta_t gt dt_det_gl / 2.)
                  sig_around = sqrt(median((signal(w, ib) - med_gl)^2.))
                  if (peak gt thres_gl * dev and sig_around lt thres_around * dev) then begin
                     w = where(delta_t le dt_det_gl / 2.)
                     wz = where(abs(signal(w, ib) - med_gl) gt thres_det * hfnoise_pass(wgl(i)), cz)
                     if cz ge 1 then begin
                        wz = w(min(wz):max(wz))
                        w_glitches = [w_glitches, wz + nt * ib]

                        n_visu_gl += 1L
                        if keyword_set(debug) then if (n_visu_gl ge 100 and n_visu_gl le 120) then begin
                           time_2 = systime(/seconds)
                           t_proc += (time_2 - time_1)
                           wset, 0
                           w = where(time_s ge time_s(it) - 5. and time_s le time_s(it) + 5.)
                           wdef = where(finite(signal(w, ib)) gt 0)
                           otherplot, psym=-4, time_s(w(wdef)), signal(w(wdef), ib), xtitle='time (s)', $
                              ytitle='signal (Jy/bol.)', title='bolometer ' + chain(ib)
                           otherplot, /over, color=10, psym=4, symbsize=2., time_s(wz), signal(wz, ib)
                           res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
                           time_1 = systime(/seconds)
                        endif

                     endif
                  endif
               endfor
            endif

            if keyword_set(debug) then save_npass_save(k) = npass_save



               ;;; quasi-uniform flux -> coarse grid:
;;;               if npass_save ge max([npass_crit, floor(norm_npass_cen * fact_npass_cen * npass_cen)]) then begin
               if (npass_save ge max([npass_crit, floor(norm_npass_cen * fact_npass_cen * npass_cen)]) $
                  or (diffuse eq 1 and npass_save ge npass_crit)) then begin
;;;
                  map_proj(wproj(k) mod dimxy_iter) = 1
;;;
if diffuse eq 1 then begin
   wsave = wcen
   npass_save = npass_cen
endif
;;;

                  ;;;wcorr = wsave
                  ;;;npass_corr = npass_save
                  wcorr = wcen
                  npass_corr = npass_cen
                  itmean_pass = wt((ind_startpass(wcorr) + ind_endpass(wcorr)) / 2)
                  it_aver = tab_aver(itmean_pass)

;;;                  weight = ns_pass(wsave) / sq_maxnoise_pass(wsave)
                  weight = 1. / sq_maxnoise_pass(wsave)
;;;                  val = total(med_pass(wsave) * weight) / total(weight)
                  val = median(med_pass(wsave), /even)
;;;
                  weight_val = weight
                  npass_val = npass_save


                  if pointcorr eq 1 then begin
                     weight_val = ns_pass / sq_maxnoise_pass
                     val = total(med_pass * weight_val) / total(weight_val)
                     npass_val = npass
                  endif

                  drift_pass = med_pass(wcorr) - val
;;;                  weight = ns_pass(wcorr) / sq_maxnoise_pass(wcorr)
                  weight = 1. / sq_maxnoise_pass(wcorr)
                  weight = 1. / (1. / weight + median(1. / weight, /even) / float(npass_val))

                  bolo_pass = wbolo(ind_startpass(wcorr))
                  ind_array = it_aver + nt_save * bolo_pass
                  array_drifts(ind_array) += (drift_pass * weight)
                  array_weight(ind_array) += weight
                  array_npass(ind_array) += (npass_corr * weight)

               endif else if flag_fine eq 1 then begin



               ;;; varying flux -> finer grid:
                  xpf0 = min(floor(x_bolos(wp)))
                  ypf0 = min(floor(y_bolos(wp)))
                  for kf = 0L, nproj_zoom - 1 do begin

                     xpf = xpf0 + tab_dxpf(kf)
                     ypf = ypf0 + tab_dypf(kf)
                     wpf = where(floor(x_bolos(wp)) eq xpf and floor(y_bolos(wp)) eq ypf, cpf)
                     if cpf ge 1 then wtf = wt(wpf) else wtf = 0
;;;                     if cpf ge 3 * mean(ntmin_pass_fine(wtf)) then begin
                     if cpf ge 3 * (min_nsamp / fact) then begin
;;;
                        wpsub = wp(wpf)
                        wfbolo = wbolo(wpf)
                        distrib_time_f = distrib_time(wpf)
                        diff_bolo = [1L, wfbolo(1:*) - wfbolo(0:cpf-2)]
                        diff_time = [0.D0, distrib_time_f(1:*) - distrib_time_f(0:cpf-2)]
                        ind_startpass = where(diff_bolo ge 1L or diff_time gt maxtimediff, npass)
                        if npass ge 2 then ind_endpass = [ind_startpass(1:*) - 1, cpf-1] else ind_endpass = [cpf-1]
                        distrib_flux_f = diff_signals(wpsub)
                        distrib_true_flux_f = signal(wpsub)

                        ns_pass = ind_endpass - ind_startpass + 1
                        flag_ntmin = intarr(npass)
                        tab_ntmin = intarr(npass)
                        for i_ntmin = 0L, npass - 1L do tab_ntmin(i_ntmin) = $
                           median(ntmin_pass_fine(wtf(ind_startpass(i_ntmin):ind_endpass(i_ntmin))), /even)
                        wcen = where(ns_pass ge tab_ntmin, npass_cen)
                        if npass_cen ge 1 then flag_ntmin(wcen) = 1

                        ws = tab_subscan(wtf(ind_startpass))
                        hfnoise_pass = highfnoise(ws, wfbolo(ind_startpass))
                        maxnoise_pass = maxnoise(ws, wfbolo(ind_startpass))


                        ;;; reduce execution time by looping on maximum number of samples per crossing
                        ;;; instead of looping on number of crossings (which scales with number of bolometers):

                        ns_max = max(ns_pass)
                        tab_flux = fltarr(ns_max, npass) + !values.f_nan
                        for ind_s = 0, ns_max - 1 do begin
                           w = where(ns_pass gt ind_s)
                           tab_flux(ind_s, w) = distrib_flux_f(ind_startpass(w)+ind_s)
                           tab_true_flux(ind_s, w) = distrib_true_flux_f(ind_startpass(w)+ind_s)
                        endfor
                        med_pass = total(tab_flux, 1, /nan) / float(ns_pass)
                        true_med_pass = total(tab_true_flux, 1, /nan) / float(ns_pass)
                        var_pass = total(abs(tab_flux - (med_pass ## (fltarr(ns_max) + 1.))), 1, /nan) / float(ns_pass)
                        true_var_pass = total(abs(tab_true_flux - (true_med_pass ## (fltarr(ns_max) + 1.))), 1, /nan) / float(ns_pass)
                        sq_maxnoise_pass = maxnoise_pass^2.

                        wsave = where(flag_ntmin eq 1 and true_var_pass lt hfnoise_pass and var_pass lt hfnoise_pass, npass_save)


                        if npass_save ge max([npass_crit, floor(norm_npass_cen * fact_npass_cen * npass_cen)]) then begin

                           map_proj(wproj(k) mod dimxy_iter) = 3

                           itmean_pass = wtf((ind_startpass(wsave) + ind_endpass(wsave)) / 2)
                           it_aver = tab_aver(itmean_pass)

;;;                           weight = ns_pass(wsave) / sq_maxnoise_pass(wsave)
                           weight = 1. / sq_maxnoise_pass(wsave)
;;;
;;;                           val = total(med_pass(wsave) * weight) / total(weight)
                  val = median(med_pass(wsave), /even)
;;;

                           drift_pass = med_pass(wsave) - val
                           weight = 1. / (1. / weight + median(1. / weight, /even) / float(npass_save))

                           bolo_pass = wfbolo(ind_startpass(wsave))
                           ind_array = it_aver + nt_save * bolo_pass
                           array_drifts(ind_array) += (drift_pass * weight)
                           array_weight(ind_array) += weight
                           array_npass(ind_array) += (npass_save * weight)
                        endif
                     endif
                  endfor
               endif
            endif
         endif
      endif
   endif
endfor

w = where(array_weight gt 0)
array_drifts(w) /= array_weight(w)
array_npass(w) /= array_weight(w)
array_weight = 0


if keyword_set(detect_gl) then begin
   nb_glitch = n_elements(w_glitches) - 1
   if nb_glitch ge 1 then begin
      w_glitches = w_glitches(1:*)
      w_glitches = w_glitches(uniq(w_glitches, sort(w_glitches)))
      nb_glitch = n_elements(w_glitches)
   endif
endif else nb_glitch = 0


;;;
;disp_ima, win=1, map_cp, title='map_cp'
;stop
;;;
if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   if keyword_set(debug) then begin
      wset, 0
      histo = histogram(save_npass_save, locations=index, min=2)
      otherplot, index, histo, psym=10, charsize=1.5, $
         title='histogram of npass_save ; npass_crit = ' + chain(npass_crit)
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif

   disp_ima, map_proj, win=5, title='drift mode: 1 -> coarse grid ; 3 -> fine grid ; 2 -> none', $
      min_map=0.01, col_tab=15, /small
   disp_ima, map_pointcorr, win=6, title='pointing correction: 1 -> not applied ; 2 -> applied or no drift', $
      min_map=0.01, col_tab=15, /small
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif

end
