pro compute_nspd_artemis, file_scalars=file_scalars, drifts_sum=drifts_sum, maxnoise=maxnoise, $
                  ind_scans=ind_scans, ind_subscans=ind_subscans, title=title, $
                  ind_sc_coarse=ind_sc_coarse


;;; computes spectral density of removed low-frequency noise (on coarse time grid)

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_times

restore, file_time_coarse


time_aver = median(time_save(1:nt_save-1) - time_save(0:nt_save-2))



nt_scan = min(ind_subscans(1:nsubscans) - ind_subscans(0:nsubscans-1))
ind_start = ind_subscans
nstart = nsubscans
nt_scan = floor(timestep * nt_scan / time_aver) + 1
power = 0
repeat begin
   nt_div = nt_scan / 2^power
   rest = nt_div mod 2
   if rest gt 0 then nt_scan -= 2^power
   power += 1
endrep until nt_scan / 2^power le 6


dt = nt_scan * time_aver
logf0 = -alog10(dt) - 0.1
freq = [findgen(nt_scan/2+1) / dt, -reverse((findgen(nt_scan/2-1) + 1.) / dt)]
nsp_bolo = fltarr(nt_scan, nb)
norm_nsp = fltarr(nt_scan, nb)


for i = 0L, nstart - 1L do begin
   w = where(time_save ge time_s(ind_start(i)))
   power = abs(fft(drifts_sum(w(0):w(0)+nt_scan-1, *), dimension=1))
   nsp_bolo += power * sqrt(dt)
   w = where(power gt 0., cw)
   if cw ge 1 then norm_nsp(w) += 1.
endfor

w = where(norm_nsp gt 0.)
nsp_bolo(w) /= norm_nsp(w)
nsp_aver = total(nsp_bolo * (norm_nsp gt 0.), 2) / total((norm_nsp gt 0.), 2)



amp_nsp = fltarr(nb)
w = where(freq gt 0.)
m = min(abs(freq(w) - 0.1), ind_01hz)
for b = 0L, nb - 1L do if total(finite(maxnoise(*, b)), 1) gt nsubscans / 4L then begin
   res = poly_fit(alog10(freq(w)), alog10(nsp_bolo(w, b)), 3, yfit=lognsp_smooth)
   amp_nsp(b) = 10.^lognsp_smooth(ind_01hz)
endif


w = where(amp_nsp gt 0.)
histo = histogram(amp_nsp(w), locations=index, binsize=median(maxnoise)*sqrt(timestep)/100.)
nbin = n_elements(histo)
histo_cumul = fltarr(nbin)
for i = 0L, nbin - 1 do histo_cumul(i) += total(histo(0:i))
histo_cumul /= max(histo_cumul)
wset, 0

w = where(histo_cumul le 0.25, cw)
if cw eq 0 then w = 0
w_inf = where(amp_nsp gt 0. and amp_nsp le max(index(w)), c_inf)
w = where(histo_cumul ge 0.75)
w_sup = where(amp_nsp ge min(index(w)), c_sup)
nsp_inf = total(reform(nsp_bolo(*, w_inf), nt_scan, c_inf), 2) / float(c_inf)
nsp_sup = total(reform(nsp_bolo(*, w_sup), nt_scan, c_sup), 2) / float(c_sup)



wset, 0
logf1 = -alog10(time_aver) - 0.2
w = where(freq gt 0., cw)
wdef = where(finite(maxnoise) gt 0)
hfnoise = mean(maxnoise(wdef)) * sqrt(timestep)
hfnoise_bolo = total(maxnoise, 1) / float(nscans) * sqrt(timestep)
yrange = [alog10(min([nsp_inf(w(0:cw-2)), hfnoise]))-0.5, alog10(max(nsp_sup(w)))+0.2]
otherplot, alog10(freq(w)), alog10(nsp_aver(w)), psym=-4, $
   charsize=2., xrange=[logf0, logf1], yrange=yrange, $
   xtitle='log f (Hz)', ytitle='log (low-f noise SD) (Jy/bol. Hz!u-0.5!n)', title=title
otherplot, /over, alog10(freq(w)), alog10(nsp_inf(w)), lstyle=5
otherplot, /over, alog10(freq(w)), alog10(nsp_sup(w)), lstyle=2
otherplot, /over, alog10(freq(w)), w*0.+alog10(hfnoise)
otherplot, /over, lstyle=5, alog10(freq(w)), w*0.+alog10(mean(hfnoise_bolo(w_inf)))
otherplot, /over, lstyle=2, alog10(freq(w)), w*0.+alog10(mean(hfnoise_bolo(w_sup)))
res = dialog_message('   Continue ?   ', /center, /information, title='Pause')


end
