pro compute_wnoise_artemis, file_scalars=file_scalars, maxnoise, highfnoise, update_mask=update_mask, $
                    update_proj=update_proj, ind_scans=ind_scans, ind_subscans=ind_subscans, $
                    initial=initial, uniformw=uniformw, mask_gl=mask_gl, f_fwhm=f_fwhm, $
                    map_back=map_back, time_1=time_1, t_proc=t_proc


;;; computes high-frequency noise estimates from spectral densities
;;; "white noise": used to define convergence criterion and compute weights
;;; "threshold noise": used to distinguish between uniform and non-uniform brightness

;;; also performs a first-pass deglitching

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

nb_line = 16L
nb_col = 18L


restore, file_signal
;;; beware: signal is modified, but the modifications must NOT be saved !

restore, file_times

tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]


if (keyword_set(debug) or keyword_set(mask_gl)) then begin
   signal_save = signal
   if keyword_set(mask_gl) then begin
      restore, file_maps
      map_signal_save = map_signal
      map_weight_save = map_weight
   endif
endif

n_gl = 0L
nexc = 0L


db = 0
w = where(finite(signal(*, bref)) gt 0, cw)
while (bref + db lt nb - 1 and cw lt nt / 3) do begin
   db += 1
   w = where(finite(signal(*, bref+db)) gt 0, cw)
endwhile
if cw lt nt / 3 then begin
   print, ''
   print, 'too many samples set to NaN !'
   print, 'Please check the mask settings.'
   print, 'Exit.'
   retall
endif
if (keyword_set(debug)) then signal_ref = signal(*, bref+db)


mask_bolo = 1. + fltarr(nt, nb)
w = where(finite(signal) lt 1, cw)
if cw ge 1 then mask_bolo(w) = 0.
w = 0

signal_interp = signal

restore, file_vscan
w = where(vscan le 0., cw)
if cw ge 1 then signal_interp(w) = !values.f_nan



;;; mask out sources before computing spectral densities:

if keyword_set(map_back) then begin
   restore, file_pixcoords
   simu_back = map_back(floor(x_bolos)+dimx*floor(y_bolos))
   for s = 0, nscans - 1 do begin
      wz = where(finite(simu_back(ind_scans(s):ind_scans(s+1)-1, *)) lt 1, cz)
      if cz ge 1 then begin
         tab = signal_interp(ind_scans(s):ind_scans(s+1)-1, *)
         tab(wz) = !values.f_nan
         signal_interp(ind_scans(s):ind_scans(s+1)-1, *) = tab
         w = where(total(finite(signal(ind_scans(s):ind_scans(s+1)-1, *)), 1) gt (ind_scans(s+1)-ind_scans(s)) / 2L, nb_valid)
         wz = where(total(finite(simu_back(ind_scans(s):ind_scans(s+1)-1, *)), 2) lt nb_valid / 2L, cz)
         if cz ge 1 then vscan(ind_scans(s)+wz) = 0.
      endif
   endfor
   x_bolos = 0
   y_bolos = 0
   simu_back = 0
endif



noise_interp = fltarr(nsubscans)
thres_interp = 12.
for s = 0L, nsubscans - 1L do begin
   diff = abs(signal_interp(ind_subscans(s)+1:ind_subscans(s+1)-1, *) - signal_interp(ind_subscans(s):ind_subscans(s+1)-2, *))
   w = where(diff gt 0., cw)
   if cw gt (ind_subscans(s+1) - ind_subscans(s)) / 4L then begin
      sig = sqrt(median(diff(w)^2.))
      if sig lt 1.e-3 then sig = sqrt(mean(diff(w)^2.))
      w = where(diff gt 0. and diff lt thres_interp * sig)
      noise_interp(s) = sqrt(mean(diff(w)^2.) / 2.)
   endif
endfor
for s = 0, nscans - 1 do begin
   wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1L)
   wz = where(noise_interp(wsubs) lt (Machar()).eps, cz, complement=wdef)
   if cz ge 1 then noise_interp(wsubs(wz)) = median(noise_interp(wsubs(wdef)), /even)
endfor

if keyword_set(visu) then begin
   wset, 0
   otherplot, lindgen(nsubscans), noise_interp, psym=4, xtitle='subscan index', ytitle='fiducial noise + replacement noise'
   for s = 0, nscans - 1 do begin
      wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1L)
      otherplot, /over, color=200, wsubs(0)+[0,0], [min(noise_interp), max(noise_interp)]
   endfor
endif


noise_interp_init = noise_interp
thres = 9. * noise_interp


if (not keyword_set(initial)) then maxnoise_save = maxnoise

maxnoise = fltarr(nsubscans, nb) + !values.f_nan
highfnoise = fltarr(nsubscans, nb) + !values.f_nan

logf_min = -1.05
dlogf = 0.1
logf_max = -alog10(2. * timestep)
nf = 1 + floor((logf_max - logf_min) / dlogf)
logf_bin = logf_min + dlogf * findgen(nf)


maxnoise_init = 1. + fltarr(nsubscans, nb)


;;; initialize seed to a constant value so that we get reproducible results:
seed = 1111L

do_subscans = 1 + intarr(nscans)
for s = 0, nscans - 1 do begin
   wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1, csubs)
   w = where(total(finite(signal(ind_scans(s):ind_scans(s+1)-1, *)), 1) gt (ind_scans(s+1) - ind_scans(s)) / 2L, nb_valid)
   wz = where(total(finite(signal_interp(ind_scans(s):ind_scans(s+1)-1, *)), 2) lt nb_valid / 2L, cz)
   tab_nt_subs = ind_subscans(wsubs(1:csubs-1)) - ind_subscans(wsubs(0:csubs-2))
   if (cz gt (ind_scans(s+1) - ind_scans(s)) / 4L or min(tab_nt_subs) lt max(tab_nt_subs) / 2.) then begin
      do_subscans(s) = 0
      for b = 0L, nb - 1L do begin
         wz = where(finite(signal_interp(ind_scans(s):ind_scans(s+1)-1, b)) lt 1, cz, complement=wdef, ncomplement=cdef)
         if cz ge 1 then begin
            if cz lt (9L * (ind_scans(s+1) - ind_scans(s))) / 10L then begin
               tab_sig = [signal_interp(ind_scans(s)+wdef(0), b), signal_interp(ind_scans(s)+wdef, b), $
                  signal_interp(ind_scans(s)+wdef(cdef-1), b)]
               tab_time = [time_s(ind_scans(s))-timestep, time_s(ind_scans(s)+wdef), time_s(ind_scans(s+1)-1)+timestep]
               signal_interp(ind_scans(s)+wz, b) = interpol(tab_sig, tab_time, time_s(ind_scans(s)+wz)) $
                  + median(noise_interp(wsubs), /even) * randomn(seed, cz)
            endif else maxnoise_init(wsubs, b) = !values.f_nan
         endif
      endfor
   endif else for subs = wsubs(0), wsubs(csubs-1) do begin
      w = where(vscan(ind_subscans(subs):ind_subscans(subs+1)-1) ge 1., cw, ncomplement=clv)
      nt_thres = clv + (3L * cw) / 4L
      for b = 0L, nb - 1L do begin
         wz = where(finite(signal_interp(ind_subscans(subs):ind_subscans(subs+1)-1, b)) lt 1, cz, complement=wdef, ncomplement=cdef)
         if cz ge 1 then begin
            if cz - clv lt (3L * cw) / 4L then begin
               tab_sig = [signal_interp(ind_subscans(subs)+wdef(0), b), signal_interp(ind_subscans(subs)+wdef, b), $
                  signal_interp(ind_subscans(subs)+wdef(cdef-1), b)]
               tab_time = [time_s(ind_subscans(subs))-timestep, time_s(ind_subscans(subs)+wdef), time_s(ind_subscans(subs+1)-1)+timestep]
               signal_interp(ind_subscans(subs)+wz, b) = interpol(tab_sig, tab_time, time_s(ind_subscans(subs)+wz)) $
                  + noise_interp(subs) * randomn(seed, cz)
            endif else maxnoise_init(subs, b) = !values.f_nan
         endif
      endfor
   endfor
endfor



if keyword_set(debug) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   wset, 0
   w = where(finite(signal_ref) gt 0)
   otherplot, time_s(w), signal_ref(w), xtitle='time (s)', ytitle='bolometer '+chain(bref), $
      title='red: after replacing missing values and samples seeing compact sources'
   otherplot, /over, color=10, time_s(w), signal_interp(w, bref+db)

   signal = signal_save - signal_interp
   for i = 0L, nsubscans - 1L do begin
      w = where(finite(maxnoise_init(i, *)) lt 1, cw)
      if cw ge 1 then signal(ind_subscans(i):ind_subscans(i+1)-1, w) = !values.f_nan
   endfor
   save, file=file_signal, signal, /xdr
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
;;;
   proj_map_nearest_artemis, file_scalars=file_scalars, maxnoise=maxnoise_init, ind_subscans=ind_subscans
;;;
   restore, file_maps
   disp_ima, map_signal, win=7, title='compact sources replaced to compute white noise', $
      weightmap=map_weight, noise_level=10.*mean(noise_interp)

   res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
   map_signal = 0
   map_weight = 0
   wdelete, 7
   time_1 = systime(/seconds)
endif



spd_bin_bolo = fltarr(nf, nb)

;;;
for s = 0, nscans - 1 do begin
   wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1L, csubs)
   if (do_subscans(s) eq 1 and csubs ge 4) then begin
      ind_scans_fft = [ind_subscans(wsubs), ind_scans(s+1)]
      nscans_fft = csubs
   endif else begin
      ind_scans_fft = ind_scans(s:s+1)
      nscans_fft = 1
   endelse
;;;

for i = 0L, nscans_fft - 1L do begin
   w = where(total(mask_bolo(ind_scans_fft(i):ind_scans_fft(i+1)-1, *), 2) gt 0., nt_scan)
   flux_scan = signal_interp(ind_scans_fft(i)+w, *)
   mask_scan = mask_bolo(ind_scans_fft(i)+w, *)

   power = 0
   repeat begin
      nt_div = nt_scan / 2^power
      rest = nt_div mod 2
      if rest gt 0 then nt_scan -= 2^power
      power += 1
   endrep until nt_scan / 2^power le 6
   dt = timestep * nt_scan
   freq = [findgen(nt_scan/2+1) / dt, -reverse((findgen(nt_scan/2-1) + 1.) / dt)]

   power_bolo = abs(fft(flux_scan(0:nt_scan-1, *), dimension=1))

   spd_bolo = power_bolo * sqrt(dt)
   wb = where(total(mask_scan) gt nt_scan / 4 and finite(total(spd_bolo, 1)) gt 0, cwb)
   spd = total(spd_bolo(*, wb), 2) / float(cwb)

   w = where(freq gt 0.)
   logf = alog10(freq(w))
   for kf = 0, nf - 1 do begin
      wbin = where(abs(logf - logf_bin(kf)) le dlogf / 2., cbin)
      if cbin ge 1 then spd_bin_bolo(kf, *) = total(spd_bolo(w(wbin), *), 1) / float(cbin)
   endfor

   w = where(logf_bin gt logf_inf and logf_bin lt logf_sup, cw)
   if cw eq 0 then begin
      print, ''
      print, 'inadapted frequency windows !'
      print, 'Use the /parallel option if the data were acquired in parallel mode.'
      print, 'Exit.'
      retall
   endif
   x = total(spd_bin_bolo(w, *), 1) / float(cw)
   if nscans_fft gt 1 then maxnoise(wsubs(0)+i, *) = x else maxnoise(wsubs, *) = (x ## (1. + fltarr(csubs)))

   w = where(logf_bin gt logf_high_inf and logf_bin lt logf_high_sup, cw)
   if cw eq 0 then begin
      print, ''
      print, 'inadapted frequency windows !'
      print, 'Use the /parallel option if the data were acquired in parallel mode.'
      print, 'Exit.'
      retall
   endif
   x = total(spd_bin_bolo(w, *), 1) / float(cw)
   if nscans_fft gt 1 then highfnoise(wsubs(0)+i, *) = x else highfnoise(wsubs, *) = (x ## (1. + fltarr(csubs)))


   if (keyword_set(debug) and keyword_set(initial) and i lt 10) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      db = 0
      while finite(maxnoise(wsubs(0L), bref+db)) lt 1 do db += 1
      wset, 0
      w = where(freq gt 0. and spd_bolo(*, bref+db) gt 0.)
      subtitle = ', scan ' + chain(s) + ', subscan ' + chain(i)
      otherplot, alog10(freq(w)), alog10(spd_bolo(w, bref+db)), title='bolometer ' + chain(bref+db) + subtitle, $
         xtitle='log f (Hz)', ytitle='log signal SD (Jy/bol. Hz!u-0.5!n)'
      otherplot, /over, color=10, alog10(freq(w)), alog10(spd(w))
      otherplot, /over, psym=4, color=200, logf_bin, alog10(spd_bin_bolo(*, bref+db))
      otherplot, /over, color=200, alog10(freq(w)), w*0. + alog10(maxnoise(wsubs(0L), bref+db)), thick=3.
      otherplot, /over, color=100, alog10(freq(w)), w*0. + alog10(highfnoise(wsubs(0L), bref+db)), thick=3.
      otherplot, /over, lstyle=2, color=200, logf_inf+fltarr(2), [-10.,10.], thick=3.
      otherplot, /over, lstyle=2, color=200, logf_sup+fltarr(2), [-10.,10.], thick=3.
      otherplot, /over, lstyle=2, color=100, logf_high_inf+fltarr(2)+0.02, [-10.,10.], thick=3.
      otherplot, /over, lstyle=2, color=100, logf_high_sup+fltarr(2)+0.02, [-10.,10.], thick=3.
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif
endfor

endfor


;;; because of FFT normalization in IDL:
maxnoise /= sqrt(timestep)
highfnoise /= sqrt(timestep)


if (not keyword_set(initial)) then begin
   w = where(finite(maxnoise_save) lt 1, cw)
   if cw ge 1 then begin
      maxnoise(w) = !values.f_nan
      highfnoise(w) = !values.f_nan
   endif
   w = where(finite(maxnoise) lt 1 and finite(maxnoise_save) gt 0, cw)
   if cw ge 1 then begin
      update_mask = 1
      update_proj = 1
   endif
endif


;;; in case the high-frequency signal is stronger than the pipeline noise amplification (SPIRE):
wdef = where(finite(maxnoise) gt 0)
ratio = highfnoise(wdef) / maxnoise(wdef)
med = median(ratio)
if med lt 1. then begin
   highfnoise_save = highfnoise
   highfnoise = maxnoise
   maxnoise = highfnoise_save
   ratio = highfnoise(wdef) / maxnoise(wdef)
   med = median(ratio)
   highfnoise(wdef) = med * maxnoise(wdef)
   flag_highf = 1
endif else flag_highf = 0

;;; in case the "threshold noise" is smaller than the "white noise":
w = where(highfnoise lt maxnoise, cw)
if cw ge 1 then highfnoise(w) = maxnoise(w)



;;; glitch masking:

if keyword_set(mask_gl) then begin
   thres_gl = 15. * median(maxnoise)
   fact_gl = 12.

   signal_glitches = fltarr(nt, nb)

   map_signal_orig = map_signal_save
   w = where(map_weight_save gt 0., complement=wz)
   map_signal_orig(wz) = !values.f_nan
   thres_weight = 0.7 * median(map_weight_save(w))

   dr_min = round(3. * fact / float(f_fwhm))
   dr_max = round(6. * fact / float(f_fwhm))
   nr = 2 * dr_max + 1
   if f_fwhm gt 4. then ns_buf = 2L else if f_fwhm gt 2. then ns_buf = 6L else ns_buf = 12L


   signal = signal_save - signal_interp
   for i = 0L, nsubscans - 1L do begin
      w = where(finite(maxnoise(i, *)) lt 1, cw)
      if cw ge 1 then signal(ind_subscans(i):ind_subscans(i+1)-1, w) = !values.f_nan
   endfor
   save, file=file_signal, signal, /xdr
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   proj_map_nearest_artemis, file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans
   restore, file_maps

   restore, file_projfine

   map_signal(0, *) = 0.
   map_signal(dimx-1, *) = 0.
   map_signal(*, 0) = 0.
   map_signal(*, dimy-1) = 0.
   wg = where(abs(map_signal) gt 0., cg)
   for ig = 0L, cg - 1 do begin
      revind_0 = revind_fine(wg(ig))
      revind_1 = revind_fine(wg(ig)+1)
      if revind_1 gt revind_0 then begin
         wp = revind_fine(revind_0:revind_1-1)
         wt = wp mod nt
         cp = revind_1 - revind_0
         ws = tab_subscan(wt)
         ws_uniq = ws(uniq(ws, sort(ws)))
         nsubs_uniq = n_elements(ws_uniq)
         ind_match_uniq = lonarr(cp)
         med_subs = !values.f_nan + fltarr(nsubs_uniq)
         ns_subs = !values.f_nan + fltarr(nsubs_uniq)
         dev_subs = !values.f_nan + fltarr(nsubs_uniq)
         for ksubs = 0L, nsubs_uniq - 1L do begin
            w = where(ws eq ws_uniq(ksubs), cw)
            ind_match_uniq(w) = ksubs
            ns_subs(ksubs) = cw
            med_subs(ksubs) = median(signal_save(wp(w)), /even)
            dev_subs(ksubs) = sqrt(mean((signal_save(wp(w)) - med_subs(ksubs))^2.))
         endfor

         med_signal = map_signal_orig(wg(ig))
         x = wg(ig) mod dimx
         y = wg(ig) / dimx
         xtot = rebin(x - dr_max + lindgen(nr), nr, nr)
         ytot = (y - dr_max + lindgen(nr)) ## (1L + lonarr(nr))
         w = where(xtot ge 0 and xtot lt dimx and ytot ge 0 and ytot lt dimy)
         wtot = xtot(w) + dimx * ytot(w)
         dr = sqrt((xtot(w) - x)^2. + (ytot(w) - y)^2.)
         w = where(dr ge dr_min and dr le dr_max)
         med_local = median(map_signal_orig(wtot(w)), /even)
         med_source = max(map_signal_orig(x-3:x+3, y-3:y+3), /nan)

         wdef = where(ns_subs ge max([3, median(ns_subs, /even) / 2.]), cdef, complement=wz, ncomplement=cz)

         ;;; glitch detection:
         if cdef ge 1 then begin
            if cdef ge 5 then begin
               wsup = where(dev_subs(wdef) ge median(dev_subs(wdef), /even))
               dev_ref = median(dev_subs(wdef(wsup)), /even)
            endif else if cdef ge 3 then dev_ref = median(dev_subs(wdef), /even) $
            else begin
               dev_ref = median(dev_subs(wdef), /even)
               if max(dev_subs(wdef), /nan) lt 1.5 * min(dev_subs(wdef), /nan) then dev_ref = 1.1 * min(dev_subs(wdef), /nan)
            endelse
            dev_ref = max([dev_ref, 1.48 * sqrt(median((med_subs(wdef) - median(med_subs(wdef), /even))^2., /even))])
         endif else dev_ref = median(maxnoise)

         if (cdef lt 3 and cz ge 1) then med_subs(wz) = med_signal $
            else if cz ge 1 then med_subs(wz) = median(med_subs(wdef), /even)
         diff_comp = max([thres_gl, fact_gl * abs(med_source - med_local), 4. * dev_ref])
         w = where(abs(signal_save(wp) - med_signal) gt diff_comp $
            and abs(signal_save(wp) - med_subs(ind_match_uniq)) gt diff_comp, cw)
         if cw ge 1 then begin
            signal_glitches(wp(w)) = signal_save(wp(w)) - min([med_subs(ind_match_uniq(w)), med_local])
            signal_save(wp(w)) = !values.f_nan
            n_gl += cw
         endif
      endif
   endfor

   print, 'glitches found by sample differences: ' + chain(n_gl) + ' sample(s) masked'

   if (keyword_set(visu) and n_gl gt 50) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      signal = signal_glitches
      for i = 0L, nsubscans - 1L do begin
         w = where(finite(maxnoise_init(i, *)) lt 1, cw)
         if cw ge 1 then signal(ind_subscans(i):ind_subscans(i+1)-1, w) = !values.f_nan
      endfor
      save, file=file_signal, signal, /xdr
      prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
      proj_map_nearest_artemis, file_scalars=file_scalars, maxnoise=maxnoise_init, ind_subscans=ind_subscans
      restore, file_maps
      disp_ima, map_signal, weightmap=map_weight, noise_level=median(maxnoise), /trunc_max, $
         win=8, title='glitches (estimate of underlying signal subtracted)'
      res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
      wdelete, 8
      time_1 = systime(/seconds)
   endif

   for b = 0L, nb - 1L do for subs = 0L, nsubscans - 1L do begin
      w = where(finite(signal_save(ind_subscans(subs):ind_subscans(subs+1)-1, b)) gt 0, cw)
      if cw le 1 then begin
         maxnoise(subs, b) = !values.f_nan
         highfnoise(subs, b) = !values.f_nan
      endif
   endfor

endif



if (keyword_set(debug) or keyword_set(mask_gl)) then begin
   signal = signal_save
   save, file=file_signal, signal, /xdr
   signal_save = 0
   signal = 0
endif

if keyword_set(debug) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   wset, 0
   for s = 0L, min([nsubscans - 1, 9]) do begin
      wdef = where(finite(maxnoise(s, *)) gt 0)
      otherplot, wdef, maxnoise(s, wdef), psym=4, xtitle='bolometer index', ytitle='white noise', $
         title='subscan '+chain(s)
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endfor
   time_1 = systime(/seconds)
endif



if keyword_set(update_mask) then begin
   restore, file_signal

   for scan = 0, nscans - 1 do begin
      nexc_min = 9999
      nexc_max = 0
      wsubs = where(ind_subscans ge ind_scans(scan) and ind_subscans lt ind_scans(scan+1) - 1L, csubs)
      ndev_unstable = !values.f_nan + fltarr(csubs, nb)

      ;;; exclude detectors that are alive only a small fraction of the time:
      wbz = where(total(finite(maxnoise(wsubs, *)), 1) gt 0 $
         and total(finite(signal(ind_scans(scan):ind_scans(scan+1)-1L, *)), 1) lt (ind_scans(scan+1) - ind_scans(scan)) / 5L, cz)
      if cz ge 1 then begin
         nexc += cz
         signal(ind_scans(scan):ind_scans(scan+1)-1L, wbz) = !values.f_nan
         for k = 0, cz - 1 do begin
            maxnoise(wsubs, wbz(k)) = !values.f_nan
            highfnoise(wsubs, wbz(k)) = !values.f_nan
         endfor
      endif

      for s = wsubs(0L), wsubs(csubs-1L) do begin
         w = where(finite(maxnoise(s, *)) lt 1, cw, ncomplement=n_valid)
         if cw ge 1 then begin
            nexc += cw
            signal(ind_subscans(s):ind_subscans(s+1)-1, w) = !values.f_nan
         endif

         ;;; exclude unstable bolometers:
         flux = signal(ind_subscans(s):ind_subscans(s+1)-1, *)
         nt_flux = ind_subscans(s+1) - ind_subscans(s)
         wb = where(total(finite(flux), 1) gt nt_flux / 2L, cb, complement=wbz, ncomplement=cz)
         if cz ge 1 then begin
            nexc += cz
            signal(ind_subscans(s):ind_subscans(s+1)-1, wbz) = !values.f_nan
            maxnoise(s, wbz) = !values.f_nan
            highfnoise(s, wbz) = !values.f_nan
         endif
         sig = flux(*, wb)
         min_sb = min(sig, /nan)
         w = where(finite(sig) lt 1, cw)
         if cw ge 1 then sig(w) = min_sb - 100.
         med_wn = median(maxnoise(s, *)) / 3.
         h2 = hist_2d(sig, findgen(cb) ## (1. + fltarr(nt_flux)), min1=min_sb, bin1=med_wn)
         n_bin = n_elements(h2(*, 0))
         sb = min_sb + med_wn * findgen(n_bin)
         m = max(h2, dimension=1, ind)
         diff = abs(sb(ind mod n_bin) - median(sig, dimension=1))
         med = median(diff)
         dev = max([med_wn/3., sqrt(mean((diff - med)^2.))])
         ndev_unstable(s-wsubs(0L), wb) = abs(diff - med) / dev

         if s eq wsubs(csubs-1L) then begin
            ndev_mean = total(ndev_unstable, 1, /nan) / total(finite(ndev_unstable), 1)
            w = where(ndev_mean gt 6., cw)
            if cw ge 1 then begin
               nexc += cw
               signal(ind_scans(scan):ind_scans(scan+1)-1, w) = !values.f_nan
               for k = 0, csubs - 1 do begin
                  maxnoise(wsubs(k), w) = !values.f_nan
                  highfnoise(wsubs(k), w) = !values.f_nan
               endfor
            endif
         endif
         if (keyword_set(debug) and s eq wsubs(csubs-1L)) then begin
            wset, 0
            wdef = where(finite(ndev_unstable(0, *)) gt 0, cdef)
            otherplot, wdef, ndev_unstable(0, wdef), psym=4, yrange=[0., 1.05*max(ndev_unstable, /nan)]
            for k = 1, csubs - 1 do begin
               wdef = where(finite(ndev_unstable(k, *)) gt 0)
               otherplot, /over, color=k*10, wdef, ndev_unstable(k,wdef), psym=4
            endfor
            if cw ge 1 then otherplot, /over, symbsize=5., w, ndev_mean(w), psym=4, xtitle='bolometer index', ytitle='ndev'
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         endif


         ;;; exclude bolometers with very deviant noise:
         wdef = where(finite(maxnoise(s, *)) gt 0)
         med_wn = median(maxnoise(s, wdef))
         sig_wn = sqrt(mean((maxnoise(s, wdef) - med_wn)^2.))
         thres = 6.
;;;
         if keyword_set(initial) then $
            w = where(maxnoise(s, *) lt max([med_wn / 5., med_wn - thres * sig_wn]), cw) $
          else w = where(maxnoise(s, *) gt min([5. * med_wn, med_wn + thres * sig_wn]) $
             or maxnoise(s, *) lt max([med_wn / 5., med_wn - thres * sig_wn]), cw)
;;;
         if (not keyword_set(initial)) then if cw gt min([n_valid / 20L, (nb_line * nb_col) / 5L]) then stop
         nexc_min = min([nexc_min, cw])
         nexc_max = max([nexc_max, cw])
;;;
         if cw ge 1 then begin
            nexc += cw
            if (keyword_set(debug) and s lt 10L) then begin
               wset, 0
               otherplot, wdef, maxnoise(s, wdef), psym=4, xtitle='bolometer index', ytitle='white noise', $
                  title='subscan ' + chain(s)
               otherplot, /over, color=10, w, maxnoise(s, w), psym=4
               res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
            endif
            signal(ind_subscans(s):ind_subscans(s+1)-1, w) = !values.f_nan
            maxnoise(s, w) = !values.f_nan
            highfnoise(s, w) = !values.f_nan
         endif

         ;;; exclude bolometers with very deviant spectral density shape:
         if flag_highf lt 1 then begin
            thres = 6.
            wdef = where(finite(maxnoise(s, *)) gt 0)
            ratio = highfnoise(s, wdef) / maxnoise(s, wdef)
            med = median(ratio)
            dev = sqrt(mean((ratio - med)^2.))
            wz = where(ratio gt max([3. * med, med + thres * dev]) $
               or ratio lt min([med / 3., med - thres * dev]), cw)
;;;
            if cw gt min([n_valid / 20L, (nb_line * nb_col) / 5L]) then stop
;;;
            if cw ge 1 then begin
               w = wdef(wz)
               nexc += cw
               if (keyword_set(debug) and s lt 10L) then begin
                  wset, 0
                  otherplot, maxnoise(s, wdef), ratio, psym=4, xtitle='white noise', $
                     ytitle='threshold noise / white noise'
                  otherplot, /over, color=10, maxnoise(s, wdef(wz)), ratio(wz), psym=4
                  res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
               endif
               signal(ind_subscans(s):ind_subscans(s+1)-1, w) = !values.f_nan
               maxnoise(s, w) = !values.f_nan
               highfnoise(s, w) = !values.f_nan
            endif
         endif
      endfor
      if nexc_max gt 0 then print, 'scan ' + chain(scan) + ' : excluding between ' + chain(nexc_min) + ' and ' + chain(nexc_max) $
         + ' noisy or dead detector(s) in ' + chain(csubs) + ' subscans'
   endfor


   db = 0
   nmax = max(total(finite(maxnoise), 1))
   while total(finite(maxnoise(*, bref-db))) lt nmax do db += 1
   if db gt 0 then begin
      bref -= db
      save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
         file_projcoarse, file_projfine, file_imprint, file_maps, $
         file_vscan, file_ntmin_pass, $
         inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
         logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
         vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
         pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
         npass_crit, fact_npass_cen, visu, debug, /xdr
   endif

   if nexc ge 1 then begin
      save, file=file_signal, signal, /xdr
      signal = 0
   endif
endif


if (keyword_set(debug) or keyword_set(mask_gl) or (keyword_set(update_mask) and nexc ge 1) or keyword_set(update_proj)) then begin
   prepare_proj_coarse, file_scalars=file_scalars
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
endif


end
