pro do_map_scans_nearest_artemis, map_scans, weight_scans, file_scalars=file_scalars, ind_scans=ind_scans, $
                  ind_subscans=ind_subscans, maxnoise=maxnoise, scan_next=scan_next, $
                  comp_errors=comp_errors, error_scans


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


restore, file_scalars

restore, file_signal

restore, file_projfine


dimxy = dimx * dimy

if (not keyword_set(scan_next)) then scan_next = 1L + lindgen(nscans)
nproj_scans = n_elements(scan_next)

map_scans = fltarr(dimx, dimy, nproj_scans)
weight_scans = fltarr(dimx, dimy, nproj_scans)
if keyword_set(comp_errors) then begin
   sqw_scans = fltarr(dimx, dimy, nproj_scans)
   error_scans = fltarr(dimx, dimy, nproj_scans)
endif



weight = 1. / maxnoise^2.
w = where(finite(weight) gt 0)
weight /= mean(weight(w))

tab_scan = lonarr(ind_scans(1))
for i = 1L, nscans - 1L do tab_scan = [temporary(tab_scan), i + lonarr(ind_scans(i+1) - ind_scans(i))]

tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]



for k = 0L, nproj_fine - 1L do begin
   revind_0 = revind_fine(wproj_fine(k))
   revind_1 = revind_fine(wproj_fine(k)+1)
   cp = revind_1 - revind_0
   wp = revind_fine(revind_0:revind_1-1)
   wt = wp mod nt
   ws = tab_scan(wt)

   bin_weight = weight(tab_subscan(wt) + nsubscans * (wp / nt))
   bin_signal = signal(wp)
   for i = 0, nproj_scans - 1 do begin
      w = where(ws eq scan_next(i) - 1, cw)
      if cw ge 1 then begin
         weight_scans(i*dimxy+wproj_fine(k)) += total(bin_weight(w))
         map_scans(i*dimxy+wproj_fine(k)) += total(bin_signal(w) * bin_weight(w))
         if keyword_set(comp_errors) then begin
            sqw_scans(i*dimxy+wproj_fine(k)) += total(bin_weight(w)^2.)
            error_scans(i*dimxy+wproj_fine(k)) += total(bin_signal(w)^2. * bin_weight(w))
         endif
      endif
   endfor
endfor

wcov = where(weight_scans gt 0.)
if keyword_set(comp_errors) then begin
   error_scans(wcov) = (error_scans(wcov) * weight_scans(wcov) - map_scans(wcov)^2.) / (weight_scans(wcov)^2. - sqw_scans(wcov))
   error_scans(wcov) = sqrt(error_scans(wcov) / weight_scans(wcov))
endif
map_scans(wcov) /= weight_scans(wcov)


end
