function proj_map_init_artemis, file_scalars=file_scalars, maxnoise=maxnoise, $
                                ind_scans=ind_scans, ind_subscans=ind_subscans


;;; project the initial map, after removing offsets (one per scan per bolometer)
;;; due to flux calibration constants

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_signal


signal_orig = signal
tab_offsets = !values.f_nan + fltarr(nscans, nb)

for scan = 0, nscans - 1 do begin
   wsubs = where(ind_subscans ge ind_scans(scan) and ind_subscans le ind_scans(scan+1) - 1)
   wb = where(total(finite(maxnoise(wsubs, *)), 1) gt 0)
   tab_offsets(scan, wb) = median(signal(ind_scans(scan):ind_scans(scan+1)-1, wb), dimension=1)
endfor


m = max(total(finite(maxnoise), 1))
wb = where(total(finite(maxnoise), 1) ge 0.9 * m)
med_offsets = median(tab_offsets(*, wb), dimension=2, /even)
offsets = total(tab_offsets - rebin(med_offsets, nscans, nb), 1, /nan) / total(finite(tab_offsets), 1)
w = where(finite(offsets) lt 1, cw)
if cw ge 1 then offsets(w) = 0.


for b = 0L, nb - 1 do signal(*, b) -= offsets(b)
save, file=file_signal, signal, /xdr

proj_map_nearest_artemis, map_init, file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans

signal = temporary(signal_orig)
save, file=file_signal, signal, /xdr


return, map_init

end
