pro proj_map_nearest_artemis, map_signal, map_weight, map_error, file_scalars=file_scalars, $
              dimx_proj=dimx_proj, dimy_proj=dimy_proj, $
              maxnoise=maxnoise, comp_errors=comp_errors, comp_lfn=comp_lfn, lowfnoise=lowfnoise, $
              todisk=todisk, ind_subscans=ind_subscans


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars


if not keyword_set(dimx_proj) then begin
   dimx_proj = dimx
   dimy_proj = dimy
endif

map_weight = fltarr(dimx_proj, dimy_proj)
map_signal = fltarr(dimx_proj, dimy_proj)
if keyword_set(comp_errors) then begin
   map_error = fltarr(dimx_proj, dimy_proj)
   map_sqw = fltarr(dimx_proj, dimy_proj)
endif else map_error = 0
if keyword_set(comp_lfn) then map_lowfnoise = fltarr(dimx_proj, dimy_proj) else map_lowfnoise = 0


weight = 1. / maxnoise^2.
w = where(finite(weight) gt 0)
weight /= mean(weight(w))


tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]


restore, file_projfine


restore, file_signal
if (keyword_set(comp_lfn) and (not keyword_set(lowfnoise))) then lowfnoise = signal * 0.


for k = 0L, nproj_fine - 1L do begin

   ;;; quicker way of evaluating: wp = where(indxy_beam eq wproj_fine(k), cp):
   revind_0 = revind_fine(wproj_fine(k))
   revind_1 = revind_fine(wproj_fine(k)+1)
   cp = revind_1 - revind_0
   wp = revind_fine(revind_0:revind_1-1)

   bin_weight = weight(nsubscans * (wp / nt) + tab_subscan(wp mod nt))
   bin_signal = signal(wp)
   map_weight(wproj_fine(k)) += total(bin_weight)
   map_signal(wproj_fine(k)) += total(bin_signal * bin_weight)

   if keyword_set(comp_errors) then begin
      map_sqw(wproj_fine(k)) += total(bin_weight^2.)
      map_error(wproj_fine(k)) += total(bin_signal^2. * bin_weight)
   endif
   if keyword_set(comp_lfn) then map_lowfnoise(wproj_fine(k)) += total(lowfnoise(wp) * bin_weight)
endfor


wcov = where(map_weight gt 0.)
if keyword_set(comp_errors) then map_error(wcov) = (map_error(wcov) * map_weight(wcov) - map_signal(wcov)^2.) $
      / (map_weight(wcov)^2. - map_sqw(wcov))
if keyword_set(comp_lfn) then map_lowfnoise(wcov) /= map_weight(wcov)
map_signal(wcov) /= map_weight(wcov)

if keyword_set(comp_errors) then map_error(wcov) = sqrt(map_error(wcov) / map_weight(wcov))


if (n_params() eq 0 or keyword_set(todisk)) then $
   save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, /xdr

end
